import numpy as np
import matplotlib.pyplot as plt


def slitsim(x0,y0,plot=False):
    """
    Simulate a star in the shutter
    """

    # in pixels units,
    # shutter is 2 x 5 pixels
    x = np.arange(-6,6,0.01)
    nx = len(x)
    y = np.arange(-10,10,0.01)
    ny = len(y)
    xx,yy = np.meshgrid(x,y)
    fwhm = 0.9
    sigma = fwhm/2.35
    x0 = -1.0 #-0.25 #0.0
    y0 = 0.0
    psf = np.exp(-0.5*((xx-x0)**2+(yy-y0)**2)/sigma**2)
    psf /= np.sum(psf)

    # shutter
    good = ((np.abs(xx)<=1) & (np.abs(yy)<=2.5))
    print('flux fraction=',np.sum(psf[good])/np.sum(psf))

    psf_masked = psf.copy()
    psf_masked[~good]=0

    xcen = np.sum(psf_masked*xx)/np.sum(psf_masked)
    ycen = np.sum(psf_masked*yy)/np.sum(psf_masked)
    print('central position=',x0,y0)
    print('boxed position=',xcen,ycen)
    print('offset=',xcen-x0,ycen-y0)

    if plot:
        pl.display(psf_masked,x,y)
        pl.oplot([-1,1,1,-1,-1],[-2.5,-2.5,2.5,2.5,-2.5],c='r')
        plt.scatter([x0],[y0],c='r',marker='+')
        plt.xlim(-2,2)
        plt.ylim(-3,3)
        plt.xlabel('X (pixels)')
        plt.ylabel('Y (pixels)')
        plt.scatter([xcen],[ycen],marker='D',facecolor='none',edgecolor='green')

def synthstar(pars,snr=100,withwiggles=True):
    """
    Create synthetic/fake star from the FERRE grid
    """

    



# Packages that allow us to get information about objects:
import asdf
import os
import numpy as np
from astropy.table import Table
from dlnpyutils import utils as dln,robust
from dlnpyutils import plotting as pl
from glob import glob
from scipy.optimize import curve_fit

# Astropy tools:
from astropy.io import fits
from doppler.spec1d import Spec1D
from scipy.ndimage import median_filter,generic_filter
from scipy.signal import argrelextrema
import doppler
from . import utils

import matplotlib
import matplotlib.pyplot as plt


# Ignore these warnings
import warnings
warnings.filterwarnings("ignore", message="OptimizeWarning: Covariance of the parameters could not be estimated")

def nanmedfilt(x,size,mode='reflect'):
    return generic_filter(x, np.nanmedian, size=size)

def relfluxcorr(spectra):
    """ Calculate the relative flux correction given spectra."""

    import pdb; pdb.set_trace()

def fluxcorr(files,outfile=None):
    """ Calculate the relative flux correction."""

    # Get the files
    files = glob(files)
    # if a single file, check if it's a list
    if len(files)==1 and utils.is_binaryfile(files[0])==False:
        listfile = files[0]
        print('Reading list file:',listfile)
        files = dln.readlines(listfile)
    if len(files)==0:
        print('No files found')
        return

    #print('KLUDGE!!!!!!!!!')
    #files = files[0:10]
    
    # Only keep files that exist and have data
    files = [f for f in files if (os.path.exists(f) and (os.path.getsize(f)>0))]
    nfiles = len(files)
    if nfiles==0:
        print('No files found that exist and have data')
        return
    print(nfiles,' files')
        
    # wavelength coefficients with linear wavelength steps
    # 3834 pixels, from 9799.765 to 18797.7624 A
    wcoef = np.array([-1.35698061e-09, -7.79636391e-06,  2.39732634e+00,  9.79971041e+03])
    npix = 3834
    xpix = np.arange(npix)
    wave = np.polyval(wcoef,xpix)

    # Load the spectra
    print('Loading the spectra')
    splist=[]
    for f in files:
        if os.path.getsize(f)>0:
            sp = doppler.read(f)
            splist.append(sp)
    nspec = len(splist)

    #------------------------------------------------    
    # Step 1.  Calculate the relative flux correction
    #            using all the spectra

    
    # Interpolate onto final grid
    spec = np.ones((nspec,npix),float)
    err = np.ones((nspec,npix),float)    
    num = np.zeros((nspec,npix),int)
    for i in range(nspec):
        sp = splist[i].copy()
        if sp.ndim==2:
            for o in range(2):
                gdf, = np.where((wave>=np.nanmin(sp.wave[:,o])) & (wave<=np.nanmax(sp.wave[:,o])))
                gd, = np.where(np.isfinite(sp.flux[:,o]) & np.isfinite(sp.wave[:,o]) &
                               (sp.mask[:,o]==False))
                if len(gd)==0 or len(gdf)==0:
                    import pdb; pdb.set_trace()
                flux1 = dln.interp(sp.wave[gd,o],sp.flux[gd,o],wave[gdf])
                err1 = dln.interp(sp.wave[gd,o],sp.err[gd,o],wave[gdf])
                # Get linear continuum fit
                coef1 = robust.polyfit(wave[gdf],flux1,1)
                cont1 = np.polyval(coef1,wave[gdf])
                # Save the normalized spectrum
                spec[i,gdf] = flux1/cont1
                err[i,gdf] = err1/cont1
                num[i,gdf] += 1
        else:
                gdf, = np.where((wave>=np.nanmin(sp.wave)) & (wave<=np.nanmax(sp.wave)))
                gd, = np.where(np.isfinite(sp.flux) & np.isfinite(sp.wave) &
                               (sp.mask==False))
                if len(gd)==0 or len(gdf)==0:
                    import pdb; pdb.set_trace()                
                flux1 = dln.interp(sp.wave[gd],sp.flux[gd],wave[gdf])
                err1 = dln.interp(sp.wave[gd],sp.err[gd],wave[gdf])
                # Get linear continuum fit
                coef1 = robust.polyfit(wave[gdf],flux1,1)
                cont1 = np.polyval(coef1,wave[gdf])
                # Save the normalized spectrum
                spec[i,gdf] = flux1/cont1
                err[i,gdf] = err1/cont1                
                num[i,gdf] += 1
                
    # Now smooth the spectra
    smspec = np.copy(spec)
    for i in range(nspec):
        #smspec[i,:] = dln.medfilt(spec[i,:],51)
        smspec[i,:] = dln.gsmooth(median_filter(spec[i,:],11),40)        

    # Throw out spectra with large scatter
    specsigma = np.zeros(nspec)+99999.
    for i in range(nspec):
        spec1 = smspec[i,:]
        err1 = err[i,:]
        gd, = np.where(np.isfinite(spec1) & np.isfinite(err1) & (err1<1e20))
        std = np.std(spec1[gd])
        sig = dln.mad(spec1[gd])
        specsigma[i] = std
    medspecsigma = np.nanmedian(specsigma)
    sigspecsigma = dln.mad(specsigma)
    gdspec, = np.where(specsigma < (medspecsigma+3*sigspecsigma))
    print('Keeping ',len(gdspec),' of ',nspec,' spectra')
    smspec = smspec[gdspec,:]
    spec = spec[gdspec,:]    
    err = err[gdspec,:]
    num = num[gdspec,:]
    nspec = len(gdspec)
    
    # Calculate the median relative offset
    reloff1 = np.nanmean(smspec,axis=0)

    #diff = np.zeros(spec.shape)
    #for i in range(nspec):
    #    diff[i,:] = smspec[i,:]-reloff
    #sig = dln.mad(diff,axis=1)
    #medsig = np.median(sig)
    #sigsig = dln.mad(sig)
    #gdspec2, = np.where(sig < medsig+3*sigsig)
    #smspec = smspec[gdspec2,:]
    #spec = spec[gdspec2,:]    
    #err = err[gdspec2,:]
    #num = num[gdspec2,:]
    #nspec = len(gdspec2)

     
    # Calculate the median relative offset
    #reloff2 = np.nanmean(smspec,axis=0)
    
    
    # fix edges
    #reloff[0:40]=0.99
    #reloff[-15:]=0.99
    
    #pl.plot(wave,reloff,yr=[0.9,1.1])
    ##pl.oplot(wave,1.0+0.03*np.sin(wave/100),c='r')
    ##pl.oplot(wave,1.0+0.015*np.sin(wave/200),c='green')
    #pl.oplot(wave,1.0+0.015*np.sin(wave/130+2.0),c='green')

    # sin(2*pi*x/lambda+phi)

    #def wavefunc(x,amp,wavelength,phi,const):
    #    return amp*np.sin(2*np.pi*x/wavelength+phi)+const
    #initpar = [0.015,820.0,2.0,1.0]
    #pars,pcov = curve_fit(wavefunc,wave,reloff,p0=initpar)
    #mm = wavefunc(wave,*pars)

    # array([ 3.72375182e-03,  7.87562008e+02, -1.25323776e+00,  1.00054443e+00])
    # array([7.45125964e-03, 8.81892624e+02, 5.23014812e+00, 9.99417155e-01])

    #import pdb; pdb.set_trace()
    
    #return wave,reloff
    
    
    # USE FLUXED CANNON MODELS to get an absolute calibration


    #---------------------------------------------------    
    # Step 2.  Apply the correction and run Doppler
    #            to get the best-fitting models for all
    #            spectra.  This is needed to remove
    #            real spectral features
    print('Running Doppler on all spectra')

    # Doppler Cannon model only covers up to 18000
    #bad = (sp.wave >= 18000)
    #if np.sum(bad)>0:
    #    sp.flux[bad] = 0.0
    #    sp.err[bad] = 1e30
    #    sp.wave[bad] = 0.0
    #    sp.mask[bad] = True

    # Loop over spectra and run Doppler
    bestmodels = []
    dt = [('vhelio',float),('vrel',float),('vrelerr',float),('teff',float),('tefferr',float),
          ('logg',float),('loggerr',float),('feh',float),('feherr',float),('chisq',float),('bc',float)]
    out = np.zeros(len(splist),dtype=np.dtype(dt))
    for i in range(len(splist)):
        sp = splist[i].copy()
        
        # Divide by the initial flux correction factor
        for j,sp1 in enumerate(sp):
            fluxcorr1 = dln.interp(wave,reloff1,sp1.wave)
            sp1.flux /= fluxcorr1
            sp1.err /= fluxcorr1
        
        # Doppler Cannon model only covers up to 18000
        bad = (sp.wave >= 17900)
        if np.sum(bad)>0:
            sp.flux[bad] = 0.0
            sp.err[bad] = 1e30
            sp.wave[bad] = 0.0
            sp.mask[bad] = True
        #    import pdb; pdb.set_trace()
        
        try:
            out1,bestmodel,spm = doppler.fit(sp,verbose=False)
            print(i+1,out1['teff'][0],out1['logg'][0],out1['feh'][0],out1['vhelio'][0])
            out[i] = out1
            bestmodels.append(bestmodel)
        except:
            print(i+1,'Doppler failed')
            bestmodels.append(None)

            
    #-------------------------------------------------------    
    # Step 3.  Divide the spectra by the best-fitting models
    #            and recalculate the relative flux correction
    
            
    # Interpolate best model onto the final grid and
    #   divide the observed spectra by it
    normspec = spec.copy()
    normerr = err.copy()    
    modelspec = np.zeros(spec.shape,float)
    for i in range(len(splist)):
        sp = splist[i].copy()        
        bestmod = bestmodels[i]
        if bestmod is None:
            continue
        for i,sp1 in enumerate(sp):
            bestmod1 = bestmod[i]
            gdf, = np.where((wave>=np.nanmin(sp1.wave)) & (wave<=np.nanmax(sp1.wave)))
            gd, = np.where(np.isfinite(sp1.flux) & np.isfinite(sp1.wave) &
                           (sp1.mask==False))
            if len(gd)==0 or len(gdf)==0:
                import pdb; pdb.set_trace()                
            flux1 = dln.interp(sp1.wave[gd],sp1.flux[gd],wave[gdf])
            err1 = dln.interp(sp1.wave[gd],sp1.err[gd],wave[gdf])            
            mgd, = np.where(np.isfinite(bestmod1.flux) & np.isfinite(bestmod1.wave) &
                           (bestmod1.mask==False))
            mflux1 = dln.interp(bestmod1.wave[mgd],bestmod1.flux[mgd],wave[gdf])
            normflux1 = flux1/mflux1
            normerr1 = err1/mflux1
            mednormflux1 = np.nanmedian(normflux1)
            normflux1 /= mednormflux1
            normerr1 /= mednormflux1
            # Save the normalized spectrum
            modelspec[i,gdf] = mflux1
            normspec[i,gdf] = normflux1
            normerr[i,gdf] = normerr1
            num[i,gdf] += 1
            

            
        #if sp.ndim==2:
        #    for o in range(2):
        #        gdf, = np.where((wave>=np.nanmin(sp.wave[:,o])) & (wave<=np.nanmax(sp.wave[:,o])))
        #        gd, = np.where(np.isfinite(sp.flux[:,o]) & np.isfinite(sp.wave[:,o]) &
        #                       (sp.mask[:,o]==False))
        #        if len(gd)==0 or len(gdf)==0:
        #            import pdb; pdb.set_trace()
        #        flux1 = dln.interp(sp.wave[gd,o],sp.flux[gd,o],wave[gdf])
        #        err1 = dln.interp(sp.wave[gd,o],sp.err[gd,o],wave[gdf])                
        #        mflux1 = dln.interp(bestmod.wave[gd,o],bestmod.flux[gd,o],wave[gdf])                
        #        normflux1 = flux1/mflux1
        #        normerr1 = err1/mflux1
        #        mednormflux1 = np.nanmedian(normflux1)
        #        normflux1 /= mednormflux1
        #        normerr1 /= mednormflux1
        #        # Get linear continuum fit
        #        #coef1 = robust.polyfit(wave[gdf],normflux1,1)
        #        #cont1 = np.polyval(coef1,wave[gdf])
        #        # Save the normalized spectrum
        #        modelspec[i,gdf] = mflux1
        #        normspec[i,gdf] = normflux1
        #        normerr[i,gdf] = normerr1
        #        num[i,gdf] += 1
        #else:
        #    gdf, = np.where((wave>=np.nanmin(sp.wave)) & (wave<=np.nanmax(sp.wave)))
        #    gd, = np.where(np.isfinite(sp.flux) & np.isfinite(sp.wave) &
        #                   (sp.mask==False))
        #    if len(gd)==0 or len(gdf)==0:
        #        import pdb; pdb.set_trace()                
        #    flux1 = dln.interp(sp.wave[gd],sp.flux[gd],wave[gdf])
        #    err1 = dln.interp(sp.wave[gd],sp.err[gd],wave[gdf])            
        #    mflux1 = dln.interp(bestmod.wave[gd],bestmod.flux[gd],wave[gdf])
        #    normflux1 = flux1/mflux1
        #    normerr1 = err1/mflux1
        #    mednormflux1 = np.nanmedian(normflux1)
        #    normflux1 /= mednormflux1
        #    normerr1 /= mednormflux1
        #    # Get linear continuum fit
        #    #coef1 = robust.polyfit(wave[gdf],normflux1,1)
        #    #cont1 = np.polyval(coef1,wave[gdf])
        #    # Save the normalized spectrum
        #    modelspec[i,gdf] = mflux1
        #    normspec[i,gdf] = normflux1
        #    normerr[i,gdf] = normerr1
        #    num[i,gdf] += 1

    gdspec, = np.where(np.nansum(normspec,axis=1) != 0)
    normspec = normspec[gdspec,:]
    normerr = normerr[gdspec,:]
    nspec = len(gdspec)
            
    # Now smooth the spectra
    smnormspec = np.zeros(normspec.shape,float)
    for i in range(nspec):
        smnormspec[i,:] = dln.gsmooth(median_filter(normspec[i,:],11),40)        

    # Throw out spectra with large scatter
    normspecsigma = np.zeros(nspec)+99999.
    for i in range(nspec):
        spec1 = smnormspec[i,:]
        err1 = normerr[i,:]
        gd, = np.where(np.isfinite(spec1) & np.isfinite(err1) & (err1<1e20))
        std = np.std(spec1[gd])
        sig = dln.mad(spec1[gd])
        normspecsigma[i] = std
    medspecsigma = np.nanmedian(normspecsigma)
    sigspecsigma = dln.mad(normspecsigma)
    gdspec, = np.where(normspecsigma < (medspecsigma+3*sigspecsigma))
    print('Keeping ',len(gdspec),' of ',nspec,' spectra')
    smnormspec = smnormspec[gdspec,:]
    normspec = normspec[gdspec,:]    
    normerr = normerr[gdspec,:]
    #num = num[gdspec,:]
    nspec = len(gdspec)
    
    # Calculate the median relative offset
    reloff = np.nanmean(smspec,axis=0)

            
    import pdb; pdb.set_trace()

    # Save to output file
    if outfile is None:
        outfile = 'nirspec_fluxcorr.fits'
    print('Writing flux correction to ',outfile)
    hdu = fits.PrimaryHDU(reloff)
    hdu.writeto(outfile,overwrite=True)
    hdu.close()




import pdb
import numpy as np

def sincint(x, nres, speclist) :
    """ Use sinc interpolation to get resampled values

        x : desired positions
        nres : number of pixels per resolution element (2=Nyquist)
        speclist : list of [quantity, variance] pairs (variance can be None)
    """

    dampfac = 3.25*nres/2.
    ksize = int(21*nres/2.)
    if ksize%2 == 0 : ksize +=1
    nhalf = ksize//2 

    #number of output and input pixels
    nx = len(x)
    nf = len(speclist[0][0])

    # integer and fractional pixel location of each output pixel
    ix = x.astype(int)
    fx = x-ix

    # outputs
    outlist=[]
    for spec in speclist :
        if spec[1] is None :
            outlist.append([np.full_like(x,0),None])
        else :
            outlist.append([np.full_like(x,0),np.full_like(x,0)])

    for i in range(len(x)) :
        xkernel = np.arange(ksize)-nhalf - fx[i]
        # in units of Nyquist
        xkernel /= (nres/2.)
        u1 = xkernel/dampfac
        u2 = np.pi*xkernel
        sinc = np.exp(-(u1**2)) * np.sin(u2) / u2
        sinc /= (nres/2.)
        
        # the value at x = 0 is defined to be the limiting value
        sinc[u2 == 0] = 1

        lobe = np.arange(ksize) - nhalf + ix[i]
        vals = np.zeros(ksize)
        vars = np.zeros(ksize)
        gd = np.where( (lobe>=0) & (lobe<nf) )[0]

        for spec,out in zip(speclist,outlist) :
            vals = spec[0][lobe[gd]]
            out[0][i] = (sinc[gd]*vals).sum()
            if spec[1] is not None : 
                var = spec[1][lobe[gd]]
                out[1][i] = (sinc[gd]**2*var).sum()

    for out in outlist :
       if out[1] is not None : out[1] = np.sqrt(out[1])
    
    return outlist



# Fit multi-band photometry with isochrones models to
# constrain Teff, A(V)

# Photometry Fitter (photter)

import numpy as np
from dlnpyutils import utils as dln
from scipy.optimize import curve_fit
from astropy.table import Table,hstack,vstack
import traceback
from . import extinction

class Fitter(object):

    # Fits stars with single metallicity isochrone
    
    def __init__(self,iso,bands,refband,extdict=None):
        self.iso = iso
        self.bands = bands
        self.nbands = len(bands)
        self.refband = refband
        metals = np.unique(iso['MH'].data)
        nmetals = len(metals)
        self.metals = metals
        self.nmetals = nmetals        
        ages = np.unique(iso['logAge'].data)
        nage = len(ages)
        self.logages = ages
        self.ages = 10**ages / 1e6   # in Myrs
        self.nages = len(ages)
        # Create an index of ages        
        if self.nages>0:
            age_index = dln.create_index(iso['logAge'].data)
            aindex = []
            for i in range(self.nages):
                aind = age_index['index'][age_index['lo'][i]:age_index['hi'][i]+1]
                aindex.append(aind)
            self.index = aindex
            self._age_index = age_index
        else:
            self.index = [np.arange(len(iso))]
        # Extinction dictionary
        if extdict is not None:
            self._extdict = extdict
        else:
            extdict,exttab = extinction.load()
            self._extdict = extdict
            self._exttab = exttab

    def __call__(self,*pars):
        """ Interpolate in the grid."""
        # Pars are teff, log(age)
        teff = pars[0]
        logteff = np.log10(teff)
        logage = pars[1]
        # Get model values

        phot = Table(np.zeros(1,dtype=self.iso.dtype))
        for c in phot.colnames: phot[c]=np.nan
        
        # Beyond ages in the model, return bad values
        if logage < np.min(self.logages) or logage >np.max(self.logages):
            return phot
            
        # At the edge
        if logage == np.min(self.logages) or logage == np.max(self.logages):
            if logage==np.min(self.logages):
                iso1 = self.iso[self.index[0]]
            else:
                iso1 = self.iso[self.index[-1]]                
            # Interpolate all iso columns
            for c in phot.colnames:
                phot[c] = dln.interp(iso1['logTe'],iso1[c],logteff,kind='quadratic',bounds_error=False,fill_value=np.nan)

        # We have two points
        else:
        
            # Get high and low ages
            aind = np.searchsorted(self.logages,logage)   # returns index just above the value
            alo = aind-1
            loage = self.logages[alo]
            loiso = self.iso[self.index[alo]]
            ahi = aind
            hiage = self.logages[ahi]
            hiiso = self.iso[self.index[ahi]]
            # wtfrac is weight of lower value, (1-wtfrac) is weight of upper value        
            wtfrac = (hiage-logage)/(hiage-loage)  

            # Interpolate all iso columns            
            for c in phot.colnames:            
                loval = dln.interp(loiso['logTe'],loiso[c],logteff,kind='quadratic',assume_sorted=False,
                                   bounds_error=False,fill_value=np.nan)
                hival = dln.interp(hiiso['logTe'],hiiso[c],logteff,kind='quadratic',assume_sorted=False,
                                   bounds_error=False,fill_value=np.nan)
                lofinite = np.isfinite(loval)
                hifinite = np.isfinite(hival)
                vals = []
                if lofinite: vals.append(loval)
                if hifinite: vals.append(hival)
                if len(vals)==0:
                    phot[c] = np.nan
                elif len(vals)==1:
                    phot[c] = vals[0]
                else:
                    phot[c] = wtfrac*vals[0] + (1-wtfrac)*vals[1]
                    
        return phot

        
    def model(self,x,*pars):
        # Pars are teff, A(V), log(age)
        teff = pars[0]
        logteff = np.log10(teff)
        av = pars[1]
        logage = pars[2]
        # Get model values

        # Beyond ages in the model, return bad values
        if logage < np.min(self.logages) or logage >np.max(self.logages):
            out = np.zeros(self.nbands)+1e30
            return out
            
        # At the edge
        if logage == np.min(self.logages) or logage == np.max(self.logages):
            if logage==np.min(self.logages):
                iso1 = self.iso[self.index[0]]
            else:
                iso1 = self.iso[self.index[-1]]                
            # Interpolate bands                
            phot = np.zeros(len(self.bands)+1,float)+99.99        
            for i,b in enumerate(self.bands+[self.refband]):
                phot[i] = dln.interp(iso1['logTe'],iso1[b],logteff,kind='quadratic',bounds_error=False,fill_value=np.nan)

        # We have two points
        else:
        
            # Get high and low ages
            aind = np.searchsorted(self.logages,logage)   # returns index just above the value
            alo = aind-1
            loage = self.logages[alo]
            loiso = self.iso[self.index[alo]]
            ahi = aind
            hiage = self.logages[ahi]
            hiiso = self.iso[self.index[ahi]]
            # wtfrac is weight of lower value, (1-wtfrac) is weight of upper value        
            wtfrac = (hiage-logage)/(hiage-loage)  
        
            lophot = np.zeros(len(self.bands)+1,float)+99.99
            hiphot = np.zeros(len(self.bands)+1,float)+99.99
            phot = np.zeros(len(self.bands)+1,float)+99.99        
            for i,b in enumerate(self.bands+[self.refband]):
                lophot[i] = dln.interp(loiso['logTe'],loiso[b],logteff,kind='quadratic',assume_sorted=False,
                                       bounds_error=False,fill_value=np.nan)
                hiphot[i] = dln.interp(hiiso['logTe'],hiiso[b],logteff,kind='quadratic',assume_sorted=False,
                                       bounds_error=False,fill_value=np.nan)
                lofinite = np.isfinite(lophot[i])
                hifinite = np.isfinite(hiphot[i])
                vals = []
                if lofinite: vals.append(lophot[i])
                if hifinite: vals.append(hiphot[i])
                if len(vals)==0:
                    phot[i] = np.nan
                elif len(vals)==1:
                    phot[i] = vals[0]
                else:
                    phot[i] = wtfrac*vals[0] + (1-wtfrac)*vals[1]
                    
        # Now apply the extinction
        data = Table()
        for i,b in enumerate(self.bands+[self.refband]):
            data[b] = [phot[i]]
        newdata = extinction.extinct(data,av,extdict=self._extdict,isonames=self.bands+[self.refband])
        
        # Subtract reference band
        colors = np.zeros(len(self.bands),float)
        refphot = newdata[self.refband]
        for i,b in enumerate(self.bands):
            colors[i] = newdata[b]-refphot
                    
        return colors

    
def fitter(data,iso,bands,refband):
    """
    data : photometry data
    iso : isochrone data
    """

    ndata = len(data)
    
    # Loop over stars
    results = []
    for i in range(ndata):
        print('--- Star {:d} ---'.format(i+1))
        besttab,sumtab = fitone(data[i],iso,bands,refband)
        print(' ')
        # Save the results
        if len(besttab)>0:
            besttab['index'] = i+1
            if len(results)==0:
                results = besttab
            else:
                results = vstack((results,besttab))

    return results


def fitone(data,iso,bands,refband):
    """
    data : photometry for single star
    iso : isochrone data
    """

    metals = np.unique(iso['MH'])
    nmetals = len(metals)
    ages = np.unique(iso['logAge'])
    nages = len(ages)

    index = dln.create_index(iso['MH'])

    dt = [('metal',float),('pars',float,3),('perrors',float,3),('chisq',float),('rms',float),('ngood',int),('success',bool)]
    sumtab = np.zeros(nmetals,dtype=np.dtype(dt))
    sumtab = Table(sumtab)

    xdata = np.zeros(len(bands),float)
    ydata = np.zeros(len(bands),float)
    errors = np.zeros(len(xdata),float)+0.05
    phdata = np.zeros(len(bands),float)
    refphot = data[refband]
    for i,b in enumerate(bands):
        phdata[i] = data[b]
    ydata = phdata - refphot
    good, = np.where((phdata > 2) & (phdata<50) & np.isfinite(phdata))
    ngood = len(good)
    if ngood==0:
        print('No good data to fit')
        return [],[]
    nbad = len(ydata)-len(good)
    if nbad>0:
        print(str(nbad)+' bad measurement(s)')
    bands1 = list(np.array(bands)[good])
    xdata = xdata[good]
    ydata = ydata[good]
    errors = errors[good]

    print('Bands: '+', '.join(bands1))
    print('Reference Band: '+str(refband))
    
    # Loop over metallicity and find the best solution
    lastpars = None
    print('  [M/H]    Teff      A(V)  log(Age)  Chisq     RMS')
    for m in range(nmetals):
        metal = index['value'][m]
        ind = index['index'][index['lo'][m]:index['hi'][m]+1]
        nind = len(ind)
        iso1 = iso[ind]

        fitter = Fitter(iso1,bands1,refband)        
        # Find best teff, AV, and log(age)
        if lastpars is None:
            estimates = [4500.0,0.1,np.median(iso['logAge'])]
        else:
            estimates = lastpars
        # physical bounds
        bounds = [np.zeros(3)-np.inf,np.zeros(3)+np.inf]
        bounds[0][0] = 10**np.min(iso['logTe'])
        bounds[1][0] = 10**np.max(iso['logTe'])        
        bounds[0][1] = 0.0
        bounds[0][2] = np.min(iso['logAge'])
        bounds[1][2] = np.max(iso['logAge'])
        try:
            pars,pcov = curve_fit(fitter.model,xdata,ydata,p0=estimates,sigma=errors,bounds=bounds)
            perrors = np.sqrt(np.diag(pcov))
            obscolors = fitter.model(xdata,*pars)
            chisq = np.sum((ydata-obscolors)**2/errors**2)
            rms = np.sqrt(np.mean((ydata-obscolors)**2))
            sumtab['metal'][m] = metal
            sumtab['pars'][m] = pars
            sumtab['perrors'][m] = perrors
            sumtab['chisq'][m] = chisq
            sumtab['rms'][m] = rms
            sumtab['ngood'][m] = ngood
            sumtab['success'][m] = True
            print('{:7.2f} {:9.3f} {:7.3f} {:7.3f} {:9.3f} {:7.3f}'.format(metal,*pars,chisq,rms))
            lastpars = pars
        except:
            traceback.print_exc()
            sumtab['metal'][m] = metal
            sumtab['ngood'][m] = ngood
            sumtab['chisq'][m] = np.nan
            sumtab['rms'][m] = np.nan            
            sumtab['success'][m] = False

    # Best value
    bestind = np.argmin(sumtab['chisq'])
    besttab = sumtab[[bestind]]
    besttab['teff'] = 0.0
    besttab['tefferr'] = 0.0    
    besttab['av'] = 0.0
    besttab['averr'] = 0.0    
    besttab['logage'] = 0.0
    besttab['logageerr'] = 0.0
    besttab['agemyr'] = 0.0    
    besttab['teff'] = besttab['pars'][0][0]
    besttab['tefferr'] = besttab['perrors'][0][0]    
    besttab['av'] = besttab['pars'][0][1]
    besttab['averr'] = besttab['perrors'][0][1]    
    besttab['logage'] = besttab['pars'][0][2]
    besttab['logageerr'] = besttab['perrors'][0][2]
    besttab['agemyr'] = (10**besttab['logage'][0])/1e6
    
    print('Best values:')
    print('{:7.2f} {:9.3f} {:7.3f} {:7.3f} {:9.3f} {:7.3f}'.format(besttab['metal'][0],*besttab['pars'][0],besttab['chisq'][0],besttab['rms'][0]))
    
    return besttab,sumtab




# Packages that allow us to get information about objects:
import asdf
import os
import numpy as np
from astropy.table import Table
from dlnpyutils import utils as dln,robust

from jwst.extract_1d.extract_1d_step import Extract1dStep
from jwst.extract_1d import extract as jextract

# Astropy tools:
from astropy.io import fits
from astropy.time import Time
from doppler.spec1d import Spec1D
from scipy.ndimage import median_filter,generic_filter
from scipy.signal import argrelextrema
from . import utils

import matplotlib
import matplotlib.pyplot as plt
from dlnpyutils import plotting as pl


# Ignore these warnings
import warnings
warnings.filterwarnings("ignore", message="OptimizeWarning: Covariance of the parameters could not be estimated")

def nanmedfilt(x,size,mode='reflect'):
    return generic_filter(x, np.nanmedian, size=size)

def findpeaks(flux):
    """ Find the peaks."""
    maxind, = argrelextrema(flux, np.greater)  # maxima
    return maxind
    
def profilefit(x,y,cenlimits=None,siglimits=None):
    """ Fit a spectral profile."""
    flux = np.sum(np.maximum(y,0))
    xmean = np.sum(x*np.maximum(y,0))/flux
    xsig = np.sqrt(np.sum((x-xmean)**2 * np.maximum(y,0)/flux))
    xsig = np.maximum(xsig,0.1)
    # Fit binned Gaussian
    p0 = [np.max(y),xmean,xsig,0.0]
    bnds = [np.array([p0[0]*0.5,xmean-1,0.5*xsig,-0.3*p0[0]]),
            np.array([p0[0]*2,xmean+1,2*xsig,0.3*p0[0]])]
    if cenlimits is not None:
        bnds[0][1] = cenlimits[0]
        bnds[1][1] = cenlimits[1]
        p0[1] = np.mean(cenlimits)
    if siglimits is not None:
        bnds[0][2] = siglimits[0]
        bnds[1][2] = siglimits[1]
        p0[2] = np.mean(siglimits)        
    if np.sum((bnds[0][:] >= bnds[1][:]))>0:
        print('problem in profilefit')
        import pdb; pdb.set_trace()

    try:
        pars,cov = dln.gaussfit(x,y,initpar=p0,bounds=bnds,binned=True)
        perror = np.sqrt(np.diag(cov))
        return pars,perror        
    except:
        print('profilefit exception')
        return None,None
    

def tracing(im,err,ytrace=None,step=15,nbin=25):
    """ Trace a spectrum. Assumed to be in the horizontal direction."""
    ny,nx = im.shape
    y,x = np.arange(ny),np.arange(nx)

    if ytrace is not None:
        if np.array(ytrace).size>1:
            ymid = np.mean(ytrace)
        else:
            ymid = ytrace
            
    # Find ymid if no trace input
    if ytrace is None:
        #tot = np.nanmedian(np.maximum(im[:,nx//2-50:nx//2+50],0),axis=1)
        tot = np.nanmedian(np.maximum(im,0),axis=1)        
        ymid = np.argmax(tot)
        
    # Trace using binned profiles, starting at middle
    nsteps = (nx//2)//step
    #if nsteps % 2 == 0: nsteps-=1
    lasty = ymid
    lastsig = 1.0
    xmnarr = []
    yhtarr = []    
    ymidarr = []
    ysigarr = []
    # Forwards
    for i in range(nsteps):
        xmn = nx//2 + i*step
        xlo = xmn - nbin//2
        xhi = xmn + nbin//2 + 1
        profile = np.nanmedian(im[:,xlo:xhi],axis=1)
        profileerr = np.nanmedian(err[:,xlo:xhi],axis=1)        
        profile[~np.isfinite(profile)] = 0.0
        profileerr[~np.isfinite(profileerr) | (profileerr<=0)] = 1e30
        flux = np.nansum(np.maximum(profile,0))
        snr = np.nanmax(profile/profileerr)  # max S/N
        if flux <= 0 or snr<5:
            continue
        ylo = np.maximum(int(np.floor(lasty-3.0*lastsig)),0)
        yhi = np.minimum(int(np.ceil(lasty+3.0*lastsig)),ny)
        slc = slice(ylo,yhi+1)
        profileclip = profile[slc]
        profileclip /= np.sum(np.maximum(profileclip,0))  # normalize
        yclip = y[slc]
        if np.sum(~np.isfinite(profileclip))>0:
            continue
        if len(yclip)==0:
            print('no pixels')
            import pdb; pdb.set_trace()
        # Limit central position using input ytrace
        #  and limit sigma
        if ytrace is not None and np.array(ytrace).size>1:
            ycen = np.mean(ytrace[xlo:xhi])
            cenlimits = [ycen-0.25,ycen+0.25]
        else:
            cenlimits = None
        siglimits = [0.4,0.6]
        pars,perror = profilefit(yclip,profileclip,cenlimits=cenlimits,siglimits=siglimits)
        if pars is None:
            continue
        xmnarr.append(xmn)
        yhtarr.append(pars[0])        
        ymidarr.append(pars[1])
        ysigarr.append(pars[2])
        # Remember
        lasty = pars[1]
        lastsig = pars[2]
        
    # Backwards
    lasty = ymid
    lastsig = 0.5
    for i in np.arange(1,nsteps):
        xmn = nx//2 - i*step
        xlo = xmn - nbin//2
        xhi = xmn + nbin//2 + 1
        profile = np.nanmedian(im[:,xlo:xhi],axis=1)
        profileerr = np.nanmedian(err[:,xlo:xhi],axis=1)        
        profile[~np.isfinite(profile)] = 0.0
        profileerr[~np.isfinite(profileerr) | (profileerr<=0)] = 1e30
        flux = np.nansum(np.maximum(profile,0))
        snr = np.nanmax(profile/profileerr)  # max S/N
        if flux <= 0 or snr<5:
            continue
        ind = np.argmax(profile)
        ylo = np.maximum(int(np.floor(lasty-3.0*lastsig)),0)
        yhi = np.minimum(int(np.ceil(lasty+3.0*lastsig)),ny)
        slc = slice(ylo,yhi+1)
        profileclip = profile[slc]
        profileclip /= np.sum(np.maximum(profileclip,0))  # normalize
        yclip = y[slc]
        if np.sum(~np.isfinite(profileclip))>0:
            continue        
        if len(yclip)==0:
            print('no pixels')
            import pdb; pdb.set_trace()
        # Limit central position using input ytrace
        #  and limit sigma
        if ytrace is not None and np.array(ytrace).size>1:
            ycen = np.mean(ytrace[xlo:xhi])
            cenlimits = [ycen-0.25,ycen+0.25]
        else:
            cenlimits = None
        siglimits = [0.4,0.6]
        pars,perror = profilefit(yclip,profileclip,cenlimits=cenlimits,siglimits=siglimits)        
        if pars is None:
            continue        
        xmnarr.append(xmn)
        yhtarr.append(pars[0])        
        ymidarr.append(pars[1])
        ysigarr.append(pars[2])
        # Remember
        lasty = pars[1]
        lastsig = pars[2]

    ttab = Table((xmnarr,ymidarr,ysigarr,yhtarr),names=['x','y','ysig','amp'])
    ttab.sort('x')
        
    return ttab

def optimalpsf(im,ytrace,err=None,off=10,backoff=50,smlen=31):
    """ Compute the PSF from the image using "optimal extraction" techniques."""
    ny,nx = im.shape
    yest = np.nanmedian(ytrace)
    # Get the subimage
    yblo = int(np.maximum(yest-backoff,0))
    ybhi = int(np.minimum(yest+backoff,ny))
    nback = ybhi-yblo
    # Background subtract
    med = np.nanmedian(im[yblo:ybhi,:],axis=0)
    medim = np.zeros(nback).reshape(-1,1) + med.reshape(1,-1)
    subim = im[yblo:ybhi,:]-medim
    suberr = imerr[yblo:ybhi,:]
    # Make sure the arrays are float64
    subim = subim.astype(float)
    suberr = suberr.astype(float)    
    # Mask other parts of the image
    ylo = ytrace-off - yblo
    yhi = ytrace+off - yblo
    yy = np.arange(nback).reshape(-1,1)+np.zeros(nx)
    mask = (yy >= ylo) & (yy <= yhi)
    sim = subim*mask
    serr = suberr*mask
    badpix = (serr <= 0)
    serr[badpix] = 1e20
    # Compute the profile/probability matrix from the image
    tot = np.nansum(np.maximum(sim,0),axis=0)
    tot[(tot<=0) | ~np.isfinite(tot)] = 1
    psf1 = np.maximum(sim,0)/tot
    psf = np.zeros(psf1.shape,float)
    for i in range(nback):
        psf[i,:] = dln.medfilt(psf1[i,:],smlen)
        #psf[i,:] = utils.gsmooth(psf1[i,:],smlen)        
    psf[(psf<0) | ~np.isfinite(psf)] = 0
    totpsf = np.nansum(psf,axis=0)
    totpsf[(totpsf<=0) | (~np.isfinite(totpsf))] = 1
    psf /= totpsf
    psf[(psf<0) | ~np.isfinite(psf)] = 0

    return psf
    
    
def extract_optimal(im,ytrace,imerr=None,verbose=False,off=10,backoff=50,smlen=31):
    """ Extract a spectrum using optimal extraction (Horne 1986)"""
    ny,nx = im.shape
    yest = np.nanmedian(ytrace)
    # Get the subo,age
    yblo = int(np.maximum(yest-backoff,0))
    ybhi = int(np.minimum(yest+backoff,ny))
    nback = ybhi-yblo
    if nback < 0:
        print('problems in extract_optimal')
        return None,None,None,None
    # Background subtract
    med = np.nanmedian(im[yblo:ybhi,:],axis=0)
    medim = np.zeros(nback).reshape(-1,1) + med.reshape(1,-1)
    subim = im[yblo:ybhi,:]-medim
    suberr = imerr[yblo:ybhi,:]
    # Make sure the arrays are float64
    subim = subim.astype(float)
    suberr = suberr.astype(float)    
    # Mask other parts of the image
    ylo = ytrace-off - yblo
    yhi = ytrace+off - yblo
    yy = np.arange(nback).reshape(-1,1)+np.zeros(nx)
    mask = (yy >= ylo) & (yy <= yhi)
    sim = subim*mask
    serr = suberr*mask
    badpix = (serr <= 0)
    serr[badpix] = 1e20
    # Compute the profile/probability matrix from the image
    tot = np.nansum(np.maximum(sim,0),axis=0)
    tot[(tot<=0) | ~np.isfinite(tot)] = 1
    psf1 = np.maximum(sim,0)/tot
    psf = np.zeros(psf1.shape,float)
    for i in range(nback):
        psf[i,:] = nanmedfilt(psf1[i,:],smlen)
        #psf[i,:] = dln.medfilt(psf1[i,:],smlen)        
        #psf[i,:] = utils.gsmooth(psf1[i,:],smlen)        
    psf[(psf<0) | ~np.isfinite(psf)] = 0
    totpsf = np.nansum(psf,axis=0)
    totpsf[(totpsf<=0) | (~np.isfinite(totpsf))] = 1
    psf /= totpsf
    psf[(psf<0) | ~np.isfinite(psf)] = 0
    # Compute the weights
    wt = psf**2/serr**2
    wt[(wt<0) | ~np.isfinite(wt)] = 0
    totwt = np.nansum(wt,axis=0)
    badcol = (totwt<=0)
    totwt[badcol] = 1
    # Compute the flux and flux error
    flux = np.nansum(psf*sim/serr**2,axis=0)/totwt
    fluxerr = np.sqrt(1/totwt)    
    fluxerr[badcol] = 1e30  # bad columns
    # Recompute the trace
    trace = np.nansum(psf*yy,axis=0)+yblo
    
    # Check for outliers
    diff = (sim-flux*psf)/serr
    bad = (diff > 25)
    if np.nansum(bad)>0:
        # Mask bad pixels
        sim[bad] = 0
        serr[bad] = 1e20
        # Recompute the flux
        wt = psf**2/serr**2
        totwt = np.nansum(wt,axis=0)
        badcol = (totwt<=0)
        totwt[badcol] = 1        
        flux = np.nansum(psf*sim/serr**2,axis=0)/totwt
        fluxerr = np.sqrt(1/totwt)
        fluxerr[badcol] = 1e30  # bad columns
        # Recompute the trace
        trace = np.nansum(psf*yy,axis=0)+yblo

    # Need at least ONE good profile point to measure a flux
    ngood = np.sum((psf>0.01)*np.isfinite(im),axis=0)
    badcol = (ngood==0)
    flux[badcol] = 0.0
    fluxerr[badcol] = 1e30 
        
    return flux,fluxerr,trace,psf


def extract_psf(im,psf,err=None,skyfit=True):
    """ Extract spectrum with a PSF."""

    if err is None:
        err = np.ones(im.shape,float)
    # Fit the sky
    if skyfit:
        # Compute the weights
        # If you are solving for flux and sky, then
        # you need to do 1/err**2 weighting
        wt = 1/err**2
        wt[(wt<0) | ~np.isfinite(wt)] = 0
        totwt = np.sum(wt,axis=0)
        badcol = (totwt<=0)
        totwt[badcol] = 1
        # Perform weighted linear regression
        flux,fluxerr,sky,skyerr = utils.weightedregression(psf,im,wt,zero=False)
        # Compute the flux and flux error
        fluxerr[badcol] = 1e30  # bad columns
        # Need at least ONE good profile point to measure a flux
        ngood = np.sum((psf>0.01)*np.isfinite(im),axis=0)
        badcol = (ngood==0)
        flux[badcol] = 0.0
        fluxerr[badcol] = 1e30
        
        return flux,fluxerr,sky,skyerr        
        
    # Only solve for flux
    #  assume sky was already subtracted
    else:
        wt = psf**2/err**2
        totwt = np.sum(wt,axis=0)
        badcol = (totwt<=0)
        totwt[badcol] = 1
        # Perform weighted linear regression
        flux,fluxerr = utils.weightedregression(psf,im,wt,zero=True)
        # Compute the flux and flux error
        fluxerr[badcol] = 1e30  # bad columns
        # Need at least ONE good profile point to measure a flux
        ngood = np.sum((psf>0.01)*np.isfinite(im),axis=0)
        badcol = (ngood==0)
        flux[badcol] = 0.0
        fluxerr[badcol] = 1e30
        
        return flux,fluxerr

def extractcol(im,err,psf):
    # Optimal extraction of a single column
    wt = psf**2/err**2
    wt[(wt<0) | ~np.isfinite(wt)] = 0
    totwt = np.nansum(wt,axis=0)
    if totwt <= 0: totwt=1
    # Compute the flux and flux error
    flux = np.nansum(psf*im/err**2,axis=0)/totwt
    fluxerr = np.sqrt(1/totwt)
    if np.isfinite(flux)==False:
        fluxerr = 1e30  # bad columns
    return flux,fluxerr

def findworstcolpix(im,err,psf):
    """ Find worst outlier pixel in a column"""
    n = len(im)
    gd = np.arange(n)    
    # Loop over the good pixels and take on away each time
    #  then recompute the flux and rchisq
    rchiarr = np.zeros(n)
    fluxarr = np.zeros(n)
    fluxerrarr = np.zeros(n)
    for j in range(n):
        ind = gd.copy()
        ind = np.delete(ind,j)
        tflx,tflxerr = extractcol(im[ind],err[ind],psf[ind])
        fluxarr[j] = tflx
        fluxerrarr[j] = tflxerr
        rchi1 = np.sum((im[ind]-tflx*psf[ind])**2/err[ind]**2)/(n-1)
        rchiarr[j] = rchi1
    bestind = np.argmin(rchiarr)
    bestrchi = rchiarr[bestind]
    bestflux = fluxarr[bestind]
    bestfluxerr = fluxerrarr[bestind]
    return bestind,bestrchi,bestflux,bestfluxerr

def fixbadpixels(im,err,psf):
    """ Fix outlier pixels using the PSF."""

    ny,nx = im.shape
    # Compute chi-squared for each column and check outliers for bad pixels
    mask = (psf > 0.01)
    # Recompute the flux
    wt = psf**2/err**2
    totwt = np.nansum(wt,axis=0)
    badcol = (totwt<=0)
    totwt[badcol] = 1        
    flux = np.nansum(psf*im/err**2,axis=0)/totwt
    # Calculate model and chisq
    model = psf*flux.reshape(1,-1)
    chisq = np.sum((model-im*mask)**2/err**2,axis=0)
    ngood = np.sum(np.isfinite(mask*im)*mask,axis=0)
    rchisq = chisq/np.maximum(ngood,1)
    #smlen = np.minimum(7,nx)
    #medrchisq = nanmedfilt(rchisq,smlen,mode='mirror')
    #sigrchisq = dln.mad(rchisq-medrchisq)
    medrchisq = np.maximum(np.nanmedian(rchisq),1.0)
    sigrchisq = dln.mad(rchisq-medrchisq,zero=True)
    coltofix, = np.where(((rchisq-medrchisq > 5*sigrchisq) | ~np.isfinite(rchisq)) & (rchisq>5) & (ngood>0))

    fixmask = np.zeros(im.shape,bool)
    fixim = im.copy()
    fixflux = flux.copy()
    fixfluxerr = flux.copy()    
    
    # Loop over columns to try to fix
    for i,c in enumerate(coltofix):
        cim = im[:,c]
        cerr = err[:,c]
        cpsf = psf[:,c]
        cflux = flux[c]
        gd, = np.where(mask[:,c]==True)
        ngd = len(gd)
        rchi = np.sum((cim[gd]-cflux*cpsf[gd])**2/cerr[gd]**2)/ngd

        # We need to try each pixel separately because if there is an
        # outlier pixel then the flux will be bad and all pixels will be "off"
        
        # While loop to fix bad pixels
        prevrchi = rchi
        count = 0
        fixed = True
        while ((fixed==True) & (ngd>2)):
            # Find worse outlier pixel in a column
            bestind,bestrchi,bestflux,bestfluxerr = findworstcolpix(cim[gd],cerr[gd],cpsf[gd])
            # Make sure it is a decent improvement
            fixed = False
            if bestrchi<0.8*prevrchi and prevrchi>5:
                curfixind = gd[bestind]
                fixmask[curfixind,c] = True
                #print('Fixing pixel ['+str(curfixind)+','+str(c)+'] ',prevrchi,bestrchi)
                # Find current and previous fixed pixels
                #  need to replace all of their values
                #  using this new flux
                fixind, = np.where(fixmask[:,c]==True)
                fixim[fixind,c] = cpsf[fixind]*bestflux
                fixflux[c] = bestflux
                fixfluxerr[c] = bestfluxerr
                gd = np.delete(gd,bestind)  # delete the pixel from the good pixel list
                ngd -= 1
                fixed = True
                prevrchi = bestrchi
            count += 1            
            
    return fixim,fixmask,fixflux,fixfluxerr

def getoffset(input_model,slit):

    im = slit.data
    
    ## Get the reference file    
    step = Extract1dStep()
    extract_ref = step.get_reference_file(input_model,'extract1d')
    extract_ref_dict = jextract.open_extract1d_ref(extract_ref, input_model.meta.exposure.type)
    
    # Get extraction parameters, from extract.create_extraction()
    slitname = slit.name
    sp_order = jextract.get_spectral_order(slit)        
    meta_source = slit
    smoothing_length = None
    use_source_posn = True
    extract_params = jextract.get_extract_parameters(extract_ref_dict,meta_source,slitname,
                                                     sp_order,input_model.meta,step.smoothing_length,
                                                     step.bkg_fit,step.bkg_order,use_source_posn)
    extract_params['dispaxis'] = extract_ref_dict['apertures'][0]['dispaxis']
    
    # Get extraction model, extract.extract_one_slit()
    # If there is an extract1d reference file (there doesn't have to be), it's in JSON format.
    extract_model = jextract.ExtractModel(input_model=input_model, slit=slit, **extract_params)
    ap = jextract.get_aperture(im.shape, extract_model.wcs, extract_params)
    extract_model.update_extraction_limits(ap)
        
    if extract_model.use_source_posn:
        #if prev_offset == OFFSET_NOT_ASSIGNED_YET:  # Only call this method for the first integration.
        offset, locn = extract_model.offset_from_offset(input_model, slit)
    
        #if offset is not None and locn is not None:
        #    log.debug(f"Computed source offset={offset:.2f}, source location={locn:.2f}")
    
        if not extract_model.use_source_posn:
            offset = 0.
        #else:
        #    offset = prev_offset
    else:
        offset = 0.
    
    extract_model.position_correction = offset
    
    # Add the source position offset to the polynomial coefficients, or shift the reference image
    # (depending on the type of reference file).
    extract_model.add_position_correction(im.shape)
    extract_model.log_extraction_parameters()
    extract_model.assign_polynomial_limits()
    
    # get disp_range, from extract.ExtractModel.extract()
    if extract_model.dispaxis == jextract.HORIZONTAL:
        slice0 = int(round(extract_model.xstart))
        slice1 = int(round(extract_model.xstop)) + 1
        #x_array = np.arange(slice0, slice1, dtype=np.float64)
        #y_array = np.empty(x_array.shape, dtype=np.float64)
        #y_array.fill((extract_model.ystart + extract_model.ystop) / 2.)
    else:
        slice0 = int(round(extract_model.ystart))
        slice1 = int(round(extract_model.ystop)) + 1
        #y_array = np.arange(slice0, slice1, dtype=np.float64)
        #x_array = np.empty(y_array.shape, dtype=np.float64)
        #x_array.fill((extract_model.xstart + extract_model.xstop) / 2.)
    disp_range = [slice0, slice1]  # Range (slice) of pixel numbers in the dispersion direction.
          
    # Get the trace, from extract1d.extract1d()
    p_src = extract_model.p_src
    srclim = []                 # this will be a list of lists, like p_src
    n_srclim = len(p_src)
    
    for i in range(n_srclim):
        lower = p_src[i][0]
        upper = p_src[i][1]
        if extract_model.independent_var.startswith("wavelength"):    # OK if 'wavelengths'
            srclim.append([lower(lambdas), upper(lambdas)])
        else:
            # Temporary array for the independent variable.
            pixels = np.arange(disp_range[0], disp_range[1], dtype=np.float64)
            srclim.append([lower(pixels), upper(pixels)])

    return offset

def fix_outliers(im,err=None,nsigma=5,nfilter=11,niter=3):
    """ Fix large outlier pixels in an image."""
    ny,nx = im.shape
    outim = im.copy()
    if err is not None:
        outerr = err.copy()
        mederr = np.nanmedian(err[im!=0])
        if np.isfinite(mederr)==0:
            mederr = np.nanmedian(err)
        outerr[~np.isfinite(outerr)] = mederr
    med = np.nanmedian(im[im!=0])
    if np.isfinite(med)==0:
        med = 0.0
    outim[~np.isfinite(outim)] = med
    for i in range(niter):
        for j in range(ny):
            line = outim[j,:]
            nonzero = (line != 0.0)
            if np.sum(nonzero)==0:
                continue        
            filt = median_filter(line,nfilter)        
            diff = line-filt
            sig = dln.mad(diff[nonzero])
            bad = (np.abs(diff) > nsigma*sig)
            if np.sum(bad)>0:
                outim[j,bad] = filt[bad]
                outerr[j,bad] *= 10  # increase the err by 10
    if err is not None:
        return outim,outerr
    else:
        return outim
        

def extract_slit(input_model,slit,backslit=None,ocalhdu=False,verbose=False,
                 applyslitcorr=True,plotbase='extract'):
    """ Extract one slit."""

    print('source_name:',slit.source_name)
    print('source_id:',slit.source_id)
    print('slitlet_id:',slit.slitlet_id)
    print('source ra/dec:',slit.source_ra,slit.source_dec)

    # Clip the X ends
    #  sometimes there are fully masked columns at the ends
    goodpix, = np.where((np.sum(np.isfinite(slit.data),axis=0)>0) & (np.sum(slit.err>0,axis=0)>0) &
                        (np.sum(np.isfinite(slit.wavelength),axis=0)>0))
    if len(goodpix)==0:
        print('No good pixels')
        return None
    xlo = goodpix[0]
    xhi = goodpix[-1]
    xstart = slit.xstart+xlo
    xsize = xhi-xlo+1
    if xlo>0:
        print('Trimming first '+str(xlo)+' fully masked columns')
    if xhi < (slit.xsize-1):
        print('Trimming last '+str((slit.xsize-1)-xhi)+' fully masked columns')
    ystart = slit.ystart
    ysize = slit.ysize    
    ystop = ystart+ysize-1
    
    # Get the data
    im = slit.data.copy()[:,xlo:xhi+1].astype(float)
    err = slit.err.copy()[:,xlo:xhi+1].astype(float)
    wave = slit.wavelength[:,xlo:xhi+1].astype(float)
    ny,nx = im.shape
    bad = (err<=0) | (~np.isfinite(err)) | (~np.isfinite(im))
    im[bad] = np.nan
    err[bad] = 1e30
    # Number of good pixels per column
    ngood = np.sum(~bad,axis=0)
    
    # Extend the wavelength information across the full image
    #  normally it only covers one section
    wave0 = wave.copy()
    y = np.arange(ny)
    ngoodwave = np.sum(np.isfinite(wave),axis=0)
    wcoef = np.zeros([nx,2],float)
    for i in range(nx):
        if ngoodwave[i]>2:
            w = wave[:,i]            
            good = np.isfinite(w)
            bad = ~np.isfinite(w)
            wcoef1 = np.polyfit(y[good],w[good],1)
            wcoef[i,:] = wcoef1
            wave[bad,i] = np.polyval(wcoef1,y[bad])
    
    if np.sum(ngood)==0:
        print('No data to extract')
        return None

    # Get the offset
    offset = getoffset(input_model,slit)

    ## Polynomial coefficients for traces
    ## y(1024) = 0
    #coef0 = np.array([ 4.59440169e-06, -1.86120908e-04, -4.83294422e+00])
    #
    ## Get peak at X=1024
    #tot = np.nansum(np.maximum(im,0),axis=1)
    #yind = np.argmax(tot)
    ## Trace for this star
    #coef = coef0.copy()
    #coef[2] += yind
    #x = np.arange(nx)
    #ytrace = np.polyval(coef,x)

    print('offset: ',offset)

    # Use the masked region from the original _cal.fits file to narrow down
    # where the trace can be
    #origim = ocalhdu.data
    #origxstart = ocalhdu.header['xstart']
    #origxsize = ocalhdu.header['xsize']
    #origystart = ocalhdu.header['ystart']
    #origysize = ocalhdu.header['ysize']    
    #omask = None

    # This returns the trace, but is offset in y by several pixels
    tr = slit.meta.wcs.get_transform('slit_frame', 'detector')
    lam = np.linspace(slit.meta.wcsinfo.waverange_start,slit.meta.wcsinfo.waverange_end,5000)
    xtr,ytr = tr(slit.source_xpos+np.zeros(5000),slit.source_ypos+np.zeros(5000),lam)
    xtr -= xlo  # correct for any trimming that we did
    gtr, = np.where((xtr >= 0) & (xtr <= (nx-1)) & (ytr >= 0) & (ytr <= (ny-1)))
    xtr = xtr[gtr]
    ytr = ytr[gtr]
    trcoef = np.polyfit(xtr,ytr,3)    
    xtrace0 = np.arange(nx)
    ytrace0 = np.interp(xtrace0,xtr,ytr)
    
    # Build Gaussian PSF model based on this trace
    y = np.arange(ny)
    yy = y.reshape(-1,1) + np.zeros(nx).reshape(1,-1)
    gmask = utils.gauss2dbin(yy,np.ones(nx),ytrace0,ytrace0*0+0.5)
    #gmask = np.sqrt(gmask)  # widen it a bit
    gmask /= np.sum(gmask,axis=0)  # normalize
    # Cross-correlation with image +/- 5 pixels
    fim,ferr = fix_outliers(im,err)  # remove outliers
    cc = np.zeros(11)
    ccerr = np.zeros(11)    
    off = np.arange(11)-5
    temperr = err.copy()
    temperr[temperr>1e20] = 0.0  # set masked pixel errors to zero
    temperr[temperr<=0] = np.nanmedian(temperr[temperr>0])
    bad = (np.abs(temperr-np.median(temperr)) > 20*dln.mad(temperr))  # get rid of really high errors
    temperr[bad] = np.median(temperr)
    for i,sh in enumerate(off):
        #cc[i] = np.sum(np.maximum(fim,0)*np.roll(gmask,sh,axis=0))
        cc[i] = np.sum(fim*np.roll(gmask,sh,axis=0))        
        ccerr[i] = np.sqrt(np.sum((temperr*np.roll(gmask,sh,axis=0))**2))
    # Find peaks
    fineoff = np.linspace(off[0],off[-1]+1,1000)
    finecc = dln.interp(off,cc,fineoff,kind='quadratic')
    fineccerr = dln.interp(off,ccerr,fineoff,kind='quadratic')    
    # Find peaks    
    pkind = findpeaks(finecc)
    ccsnr = finecc/fineccerr  # this should be the S/N of the entire spectrum
    # Only keep peaks with S/N>50
    gdpk, = np.where(ccsnr[pkind]>50)
    if len(gdpk)==0:
        print('No good CC peaks with S/N>50')
        return None
    pkind = pkind[gdpk]
    # If multiple peaks, pick the one closest to the initial trace
    if len(gdpk)>1:
        print('Multiple CC peaks.  Picking the one closest to the initial trace')
        neighbor = True
        clsind = np.argmin(np.abs(fineoff[pkind]))
        bestind = pkind[clsind]
        bestoff = fineoff[bestind]
        peakccsnr = ccsnr[bestind]
    # Only one good peak
    else:
        neighbor = False
        bestind = pkind[0]
        bestoff = fineoff[bestind]
        peakccsnr = ccsnr[bestind]
    print('Best peak CC S/N:',peakccsnr)              
    print('Cross-correlation Y offset:',bestoff)
    
    # Centroid
    imask = (gmask > 0.001)
    dyy = yy-(ytrace0.reshape(1,-1) + np.zeros(ny).reshape(-1,1))
    doff = np.sum(gmask*np.maximum(fim,0)*dyy)/np.sum(gmask*np.maximum(fim,0))
    print('Centroid Y offset:',doff)

    #yoffset = np.mean([bestoff,doff])
    yoffset = bestoff
    print('Y offset: ',yoffset)
    
    # Offset the Y trace
    ytrace1 = ytrace0 + yoffset
    
    # Do it in chunks of 50 pixels
    #nchunks = nx//100
    #doffchunks = np.zeros(nchunks)
    #for i in range(nchunks):
    #    x1 = 100*i
    #    x2 = x1+100
    #    slc1 = (slice(0,ny),slice(x1,x2))
    #    doffchunks[i] = np.sum(gmask[slc1]*np.maximum(fim[slc1],0)*dyy[slc1])/np.sum(gmask[slc1]*np.maximum(fim[slc1],0))
    
    # I think the trace is at
    # ny/2+offset at Y=1024
    #yind = (ny-1)/2+offset
    
    # Get the trace
    x = np.arange(nx)
    #ttab = tracing(im,err,yind)
    ttab = tracing(fim,ferr,ytrace1)
    
    if len(ttab)==0:
        print('Problem - no trace found')
        return None

    try:
        if len(ttab)<3:
            tcoef = np.array([np.median(ttab['y'])])
            tsigcoef = np.array([np.median(ttab['ysig'])])            
        else:
            tcoef = robust.polyfit(ttab['x'],ttab['y'],2)
            tsigcoef = robust.polyfit(ttab['x'],ttab['ysig'],1)            
        ytrace = np.polyval(tcoef,x)
        ysig = np.polyval(tsigcoef,x)
    except:
        print('tracing coefficient problem')
        import pdb; pdb.set_trace()
        
    # Create the mask
    ybin = 3
    yy = np.arange(ny).reshape(-1,1) + np.zeros(nx).reshape(1,-1)
    mask = ((yy >= (ytrace-ybin)) & (yy <= (ytrace+ybin)))


    # Get the background from the background image
    if backslit is not None:
        bim = backslit.data.copy()
        backim = np.zeros(im.shape,float)+np.nan
        bxstart = backslit.xstart
        bxsize = backslit.xsize
        bystart = backslit.ystart
        bysize = backslit.ysize
        # Find the intersection of their rows        
        # X-values
        xs = np.arange(xsize)+xstart
        backxs = np.arange(bxsize)+bxstart
        xinter = np.intersect1d(xs,backxs)
        xinter0,xinter1 = np.min(xinter),np.max(xinter)
        xlo = xinter0-xstart
        xhi = xinter1-xstart
        bxlo = xinter0-bxstart
        bxhi = xinter1-bxstart
        # Y-values
        ys = np.arange(ysize)+ystart
        backys = np.arange(bysize)+bystart
        yinter = np.intersect1d(ys,backys)
        yinter0,yinter1 = np.min(yinter),np.max(yinter)
        ylo = yinter0-ystart
        yhi = yinter1-ystart
        bylo = yinter0-bystart
        byhi = yinter1-bystart
        backim[ylo:yhi+1,xlo:xhi+1] = bim[bylo:byhi+1,bxlo:bxhi+1]
        # Multiply by the mask
        temp1 = backim.copy()
        temp1[~mask] = np.nan
        sky1 = np.nanmedian(temp1,axis=0)
        diff1 = backim.copy()-sky1.reshape(1,-1)
        sig1 = dln.mad(diff1[mask])
        temp2 = backim.copy()
        temp2[(~mask) | (np.abs(diff1)>3*sig1)] = np.nan
        sky2 = np.nanmean(temp2,axis=0)
        sky = np.zeros(ny).reshape(-1,1) + sky2.reshape(1,-1)
    else:
        # Determine a local sky
        temp1 = im.copy()
        smask = ((yy >= (ytrace-10)) & (yy <= (ytrace+10)) &
                 ((yy < (ytrace-2)) | (yy > (ytrace+2))))
        temp1[~smask] = np.nan
        sky1 = np.nanmedian(temp1,axis=0)
        diff1 = im.copy()-sky1.reshape(1,-1)
        sig1 = dln.mad(diff1[smask])
        temp2 = im.copy()
        temp2[(~smask) | (np.abs(diff1)>3*sig1)] = np.nan
        sky2 = np.nanmean(temp2,axis=0)
        sky = np.zeros(ny).reshape(-1,1) + sky2.reshape(1,-1)
    im -= sky   # subtract background

    # Build Gaussian PSF model
    y = np.arange(ny)
    yy = y.reshape(-1,1) + np.zeros(nx).reshape(1,-1)
    gpsf = utils.gauss2dbin(yy,np.ones(nx),ytrace,ysig)
    gpsf /= np.sum(gpsf,axis=0)  # normalize
    # Gaussian PSF extraction
    gflux,gfluxerr,sky,skyerr = extract_psf(im*mask,gpsf,err,skyfit=True)

    # Number of good pixels per column with good PSF
    ngood = np.sum((gpsf>0.01)*np.isfinite(im),axis=0)

    if np.sum(ngood)==0:
        import pdb; pdb.set_trace()
        print('no good columns')
        return None
    
    # Optimal extraction
    #  Use the fixed image (with fix_outliers) for the first attempt
    omask = mask.copy()
    if neighbor:
        omask = (gpsf>0.0001)   # use narrower mask (~4-5 pixels) if there is a close neighbor
    oflux1,ofluxerr1,otrace1,opsf1 = extract_optimal(fim*omask,ytrace,imerr=ferr)
    if oflux1 is None:
        return None
    
    # Fix bad pixels using the optimal PSF
    fixim,fixmask,fixflux,fixfluxerr = fixbadpixels(im,err,opsf1)

    # REJECT BAD PIXELS and redo the optimal extraction
    oflux,ofluxerr,otrace,opsf = extract_optimal(fixim*omask,ytrace,imerr=err)

    # Boxcar extraction of the fixed image
    boxflux = np.nansum(omask*fixim,axis=0)

    # Optimal extraction looks a little bit better than the boxcar extraction
    #  lower scatter
    flux = oflux
    fluxerr = ofluxerr
    
    # Get the wavelengths
    pmask = (gpsf > 0.01)
    #wav = np.nansum(wave*pmask,axis=0)/np.sum(pmask,axis=0) * 1e4  # convert to Angstroms
    #wav = np.nansum(slwave*slgpsf,axis=0)/np.sum(slgpsf*np.isfinite(slwave*slgpsf),axis=0) * 1e4  # convert to Angstroms

    # Save some diagnostic plots    
    matplotlib.use('Agg')
    fig = plt.figure(figsize=(12,7))
    medflux = np.nanmedian(oflux)
    vmin = -medflux/3.
    vmax = medflux
    pl.display(im,xtitle='X',ytitle='Y',vmin=vmin,vmax=vmax,title='DATA - SLIT_ID '+str(slit.source_id))
    pl.oplot(ytrace,c='red')
    plt.savefig(plotbase+'_data.png',bbox_inches='tight')
    pl.display(opsf,xtitle='X',ytitle='Y',vmin=0,vmax=1.2*np.max(opsf),title='Optimal PSF - SLIT_ID '+str(slit.source_id))
    pl.oplot(ytrace,c='red')
    plt.savefig(plotbase+'_opsf.png',bbox_inches='tight')
    # Flux
    plt.clf()
    plt.plot(oflux,label='Optimal',linewidth=2)
    plt.plot(gflux,label='Gaussian')
    plt.plot(boxflux,label='Boxcar',linestyle='dashed')
    plt.ylim(-medflux/3.,1.8*medflux)
    plt.xlabel('X')
    plt.ylabel('Flux')
    plt.legend()
    plt.savefig(plotbase+'_flux.png',bbox_inches='tight')
    matplotlib.use('MacOSX')

   
    # Get the wavelengths
    pmask = (opsf > 0.01)
    #wav = np.nansum(wave*pmask,axis=0)/np.sum(pmask,axis=0) * 1e4  # convert to Angstroms
    owav = np.nansum(wave*opsf,axis=0)/np.sum(opsf*np.isfinite(wave*opsf),axis=0) * 1e4  # convert to Angstroms

    # Slit offsets
    srcxpos = slit.source_xpos
    srcypos = slit.source_ypos
    
    # Get wavelengths using the trace and WCS object
    xw = x + xlo
    yw = otrace
    rr,dd,wcs_wl = slit.meta.wcs(xw,yw)
    wav = wcs_wl * 1e4
    if applyslitcorr:
        print('Applying slit correction (X,Y): (%.2f,%.2f) pixels' % (2*srcxpos,2*srcypos))        
        xw = x + xlo + 2*srcxpos
        yw = otrace + 2*srcypos
        rr,dd,wcs_wl_slitcorr = slit.meta.wcs(xw,yw)
        wav = wcs_wl_slitcorr * 1e4
    dwave = np.gradient(wav)

    #import pdb; pdb.set_trace()
    
    # Apply slit correction
    # SLIT correction, srcxpos is source position in slit
    # the slit is 2 pixels wide
    #dwave = np.gradient(wav)
    #import pdb; pdb.set_trace()
    #if applyslitcorr:
    #    newwav = wav+2*srcxpos*dwave
    #    print('Applying slit correction: %.2f pixels' % (2*srcxpos))
    #else:
    #    newwav = wav
        
    # Add the LSF information
    #  we are essentially working in a slit-less spectrograph regime
    #  the LSF is set by the seeing
    #wsig = ysig*dwave
    wsig = np.median(ysig)*dwave    
    gdw, = np.where(np.isfinite(wav) & np.isfinite(wsig))
    if len(gdw)==0:
        import pdb; pdb.set_trace()
    #import pdb; pdb.set_trace()
    #if len(gdw) > 500:
    #    wsigcoef = np.polyfit(wav[gdw],wsig[gdw],1)
    #else:
    #    wsigcoef = np.polyfit(wav[gdw],wsig[gdw],0)
    wsigcoef = np.polyfit(wav[gdw],wsig[gdw],1)        
    print('LSF sigma coefficients:',wsigcoef)
        
    # Put it all together
    sp = Spec1D(flux,err=fluxerr,wave=wav,mask=(fluxerr>1e20),instrument='NIRSpec',
                  lsfpars=wsigcoef[::-1],lsftype='Gaussian',lsfxtype='wave')
    sp.date = input_model.meta.date
    #sp.jd = Time(input_model.meta.date).jd
    sp.jd = Time(slit.meta.time.barycentric_expmid,format='mjd').jd
    sp.exptime = slit.meta.exposure.effective_exposure_time
    sp.bc = slit.meta.wcsinfo.velosys/1e3   # BC in km/s
    sp.ytrace = ytrace
    sp.source_name = slit.source_name
    sp.source_id = slit.source_id
    sp.slitlet_id = slit.slitlet_id
    sp.source_ra = slit.source_ra
    sp.source_dec = slit.source_dec
    sp.xstart = xstart
    sp.xsize = xsize
    sp.ystart = slit.ystart
    sp.ysize = slit.ysize
    sp.offset = yoffset
    sp.tcoef = tcoef
    sp.tsigcoef = tsigcoef
    
    return sp


def extract_slit_multi(input_model,slit,ratehdu=None,bratehdu=None,verbose=False,plotbase='extract'):
    """ Extract one slit."""

    print('source_name:',slit.source_name)
    print('source_id:',slit.source_id)
    print('slitlet_id:',slit.slitlet_id)
    print('source ra/dec:',slit.source_ra,slit.source_dec)
    print('ny/nx:',*slit.data.shape)

    # Clip the X ends
    #  sometimes there are fully masked columns at the ends
    goodpix, = np.where((np.sum(np.isfinite(slit.data),axis=0)>0) & (np.sum(slit.err>=0,axis=0)>0))
    if len(goodpix)==0:
        print('No good pixels')
        return None,None
    xlo = goodpix[0]
    xhi = goodpix[-1]
    xstart = slit.xstart+xlo
    xsize = xhi-xlo+1
    if xlo>0:
        print('Trimming first '+str(xlo)+' fully masked columns')
    if xhi < (slit.xsize-1):
        print('Trimming last '+str((slit.xsize-1)-xhi)+' fully masked columns')
    
    # Get the data
    slim = slit.data.copy()[:,xlo:xhi+1]
    slerr = slit.err.copy()[:,xlo:xhi+1]
    slwave = slit.wavelength[:,xlo:xhi+1]
    ny,nx = slim.shape
    slbad = (slerr<=0)
    slim[slbad] = np.nan
    slerr[slbad] = 1e30
    # Number of good pixels per column
    nslgood = np.sum(~slbad,axis=0)
    
    # Extend the wavelength information across the full image
    #  normally it only covers one section
    wave = slwave.copy()
    y = np.arange(ny)
    ngoodwave = np.sum(np.isfinite(wave),axis=0)
    wcoef = np.zeros([nx,2],float)
    for i in range(nx):
        if ngoodwave[i]>2:
            w = wave[:,i]            
            good = np.isfinite(w)
            bad = ~np.isfinite(w)
            wcoef1 = np.polyfit(y[good],w[good],1)
            wcoef[i,:] = wcoef1
            wave[bad,i] = np.polyval(wcoef1,y[bad])


    # NEED to mask bad pixels in rate images better!!!!!
    # Use the DQ information!!!
            
    # Get the "raw" rate data
    if ratehdu is not None:
        im = ratehdu[1].data.copy()
        err = ratehdu[2].data.copy()
        bad = (err<=0)
        im[bad] = np.nan
        err[bad] = 1e30
    # Get the "raw" rate background data
    if bratehdu is not None:
        bim = bratehdu[1].data.copy()
        berr = bratehdu[2].data.copy()
        bbad = (berr<=0)
        bim[bbad] = np.nan
        berr[bbad] = 1e30
        
    if np.sum(nslgood)==0:
        print('No data to extract')
        return None,None
    
    ## Get the reference file
    if True:
        step = Extract1dStep()
        extract_ref = step.get_reference_file(input_model,'extract1d')
        extract_ref_dict = jextract.open_extract1d_ref(extract_ref, input_model.meta.exposure.type)
    
        # Get extraction parameters, from extract.create_extraction()
        slitname = slit.name
        sp_order = jextract.get_spectral_order(slit)        
        meta_source = slit
        smoothing_length = None
        use_source_posn = True
        extract_params = jextract.get_extract_parameters(extract_ref_dict,meta_source,slitname,
                                                         sp_order,input_model.meta,step.smoothing_length,
                                                         step.bkg_fit,step.bkg_order,use_source_posn)
        extract_params['dispaxis'] = extract_ref_dict['apertures'][0]['dispaxis']
    
        # Get extraction model, extract.extract_one_slit()
        # If there is an extract1d reference file (there doesn't have to be), it's in JSON format.
        extract_model = jextract.ExtractModel(input_model=input_model, slit=slit, **extract_params)
        ap = jextract.get_aperture(slim.shape, extract_model.wcs, extract_params)
        extract_model.update_extraction_limits(ap)
        
        if extract_model.use_source_posn:
            #if prev_offset == OFFSET_NOT_ASSIGNED_YET:  # Only call this method for the first integration.
            offset, locn = extract_model.offset_from_offset(input_model, slit)
    
            #if offset is not None and locn is not None:
            #    log.debug(f"Computed source offset={offset:.2f}, source location={locn:.2f}")
    
            if not extract_model.use_source_posn:
                offset = 0.
            #else:
            #    offset = prev_offset
        else:
            offset = 0.
    
        extract_model.position_correction = offset
    
        # Add the source position offset to the polynomial coefficients, or shift the reference image
        # (depending on the type of reference file).
        extract_model.add_position_correction(slim.shape)
        extract_model.log_extraction_parameters()
        extract_model.assign_polynomial_limits()
    
        # get disp_range, from extract.ExtractModel.extract()
        if extract_model.dispaxis == jextract.HORIZONTAL:
            slice0 = int(round(extract_model.xstart))
            slice1 = int(round(extract_model.xstop)) + 1
            #x_array = np.arange(slice0, slice1, dtype=np.float64)
            #y_array = np.empty(x_array.shape, dtype=np.float64)
            #y_array.fill((extract_model.ystart + extract_model.ystop) / 2.)
        else:
            slice0 = int(round(extract_model.ystart))
            slice1 = int(round(extract_model.ystop)) + 1
            #y_array = np.arange(slice0, slice1, dtype=np.float64)
            #x_array = np.empty(y_array.shape, dtype=np.float64)
            #x_array.fill((extract_model.xstart + extract_model.xstop) / 2.)
        disp_range = [slice0, slice1]  # Range (slice) of pixel numbers in the dispersion direction.
          
        # Get the trace, from extract1d.extract1d()
        p_src = extract_model.p_src
        srclim = []                 # this will be a list of lists, like p_src
        n_srclim = len(p_src)
    
        for i in range(n_srclim):
            lower = p_src[i][0]
            upper = p_src[i][1]
            if extract_model.independent_var.startswith("wavelength"):    # OK if 'wavelengths'
                srclim.append([lower(lambdas), upper(lambdas)])
            else:
                # Temporary array for the independent variable.
                pixels = np.arange(disp_range[0], disp_range[1], dtype=np.float64)
                srclim.append([lower(pixels), upper(pixels)])

    ## Polynomial coefficients for traces
    ## y(1024) = 0
    #coef0 = np.array([ 4.59440169e-06, -1.86120908e-04, -4.83294422e+00])
    #
    ## Get peak at X=1024
    #tot = np.nansum(np.maximum(im,0),axis=1)
    #yind = np.argmax(tot)
    ## Trace for this star
    #coef = coef0.copy()
    #coef[2] += yind
    #x = np.arange(nx)
    #ytrace = np.polyval(coef,x)

    print('offset: ',offset)
    
    # Extract the CAL 2D SLIT IMAGE
    #------------------------------

    # I think the trace is at
    # ny/2+offset at Y=1024
    yind = (ny-1)/2+offset
    
    # Get the trace
    x = np.arange(nx)
    slttab = tracing(slim,slerr,yind)

    if len(slttab)==0:
        print('Problem - no trace found')
        return None,None

    try:
        if len(slttab)<3:
            sltcoef = np.array([np.median(slttab['y'])])
            sltsigcoef = np.array([np.median(slttab['ysig'])])            
        else:
            sltcoef = robust.polyfit(slttab['x'],slttab['y'],2)
            sltsigcoef = robust.polyfit(slttab['x'],slttab['ysig'],1)            
        slytrace = np.polyval(sltcoef,x)
        slysig = np.polyval(sltsigcoef,x)
    except:
        print('tracing coefficient problem')
        import pdb; pdb.set_trace()

        
    # Create the mask
    ybin = 3
    yy = np.arange(ny).reshape(-1,1) + np.zeros(nx).reshape(1,-1)
    slmask = ((yy >= (slytrace-ybin)) & (yy <= (slytrace+ybin)))

    # Build Gaussian PSF model
    y = np.arange(ny)
    yy = y.reshape(-1,1) + np.zeros(nx).reshape(1,-1)
    slgpsf = utils.gauss2dbin(yy,np.ones(nx),slytrace,slysig)
    slgpsf /= np.sum(slgpsf,axis=0)  # normalize
    # Gaussian PSF extraction
    slgflux,slgfluxerr,slsky,slskyerr = extract_psf(slim*slmask,slgpsf,slerr,skyfit=True)

    # Number of good pixels per column with good PSF
    slngood = np.sum((slgpsf>0.01)*np.isfinite(slim),axis=0)

    # Optimal extraction
    sloflux1,slofluxerr1,slotrace1,slopsf1 = extract_optimal(slim*slmask,slytrace,imerr=slerr)

    # Fix bad pixels using the optimal PSF
    slfixim,slfixmask,slfixflux,slfixfluxerr = fixbadpixels(slim,slerr,slopsf1)

    # REJECT BAD PIXELS and redo the optimal extraction
    sloflux,slofluxerr,slotrace,slopsf = extract_optimal(slfixim*slmask,slytrace,imerr=slerr)

    # Boxcar extraction of the fixed image
    slboxflux = np.nansum(slmask*slfixim,axis=0)

    # Optimal extraction looks a little bit better than the boxcar extraction
    #  lower scatter
    slflux = sloflux
    slfluxerr = slofluxerr
    
    # Get the wavelengths
    slpmask = (slgpsf > 0.01)
    #wav = np.nansum(wave*pmask,axis=0)/np.sum(pmask,axis=0) * 1e4  # convert to Angstroms
    #wav = np.nansum(slwave*slgpsf,axis=0)/np.sum(slgpsf*np.isfinite(slwave*slgpsf),axis=0) * 1e4  # convert to Angstroms

    # Save some diagnostic plots    
    matplotlib.use('Agg')
    fig = plt.figure(figsize=(12,7))
    medflux = np.nanmedian(sloflux)
    vmin = -medflux/3.
    vmax = medflux
    pl.display(slim,xtitle='X',ytitle='Y',vmin=vmin,vmax=vmax,title='CAL DATA - SLIT_ID '+str(slit.source_id))
    pl.oplot(slytrace,c='red')
    plt.savefig(plotbase+'_caldata.png',bbox_inches='tight')
    pl.display(slopsf,xtitle='X',ytitle='Y',vmin=0,vmax=1.2*np.max(slopsf),title='CAL Optimal PSF - SLIT_ID '+str(slit.source_id))
    pl.oplot(slytrace,c='red')
    plt.savefig(plotbase+'_calopsf.png',bbox_inches='tight')
    # Flux
    plt.clf()
    plt.plot(sloflux,label='Optimal',linewidth=2)
    plt.plot(slgflux,label='Gaussian')
    plt.plot(slboxflux,label='Boxcar',linestyle='dashed')
    plt.ylim(-medflux/3.,1.8*medflux)
    plt.xlabel('X')
    plt.ylabel('Flux')
    plt.legend()
    plt.savefig(plotbase+'_calflux.png',bbox_inches='tight')
    matplotlib.use('MacOSX')

   
    # Get the wavelengths
    slpmask = (slopsf > 0.01)
    #wav = np.nansum(wave*pmask,axis=0)/np.sum(pmask,axis=0) * 1e4  # convert to Angstroms
    slwav = np.nansum(slwave*slopsf,axis=0)/np.sum(slopsf*np.isfinite(slwave*slopsf),axis=0) * 1e4  # convert to Angstroms

    # Apply slit correction
    srcxpos = slit.source_xpos
    srcypos = slit.source_ypos
    # SLIT correction, srcxpos is source position in slit
    # the slit is 2 pixels wide
    sldwave = np.gradient(slwav)
    slnewwav = slwav+2*srcxpos*sldwave
    print('Applying slit correction: %.2f pixels' % (2*srcxpos))

    # Add the LSF information
    #  we are essentially working in a slit-less spectrograph regime
    #  the LSF is set by the seeing
    slwsig = slysig*sldwave
    gdw, = np.where(np.isfinite(slwav) & np.isfinite(slwsig))
    slwsigcoef = np.polyfit(slwav[gdw],slwsig[gdw],1)
    
    # Put it all together
    slsp = Spec1D(slflux,err=slfluxerr,wave=slwav,mask=(slfluxerr>1e20),instrument='NIRSpec',
                  lsfpars=slwsigcoef[::-1],lsftype='Gaussian',lsfxtype='wave')
    slsp.date = input_model.meta.date
    slsp.jd = Time(input_model.meta.date).jd
    slsp.ytrace = slytrace
    slsp.source_name = slit.source_name
    slsp.source_id = slit.source_id
    slsp.slitlet_id = slit.slitlet_id
    slsp.source_ra = slit.source_ra
    slsp.source_dec = slit.source_dec
    slsp.xstart = xstart
    slsp.xsize = xsize
    slsp.ystart = slit.ystart
    slsp.ysize = slit.ysize
    slsp.offset = offset
    slsp.tcoef = sltcoef
    slsp.tsigcoef = sltsigcoef

    if ratehdu is None:
        return slsp,None

    
    # Extract the RATE IMAGE with background subtraction
    #---------------------------------------------------
    
    # Now use the rate images to do the extraction
    # the XSTART/YSTART are in 1-based indexing
    slc = (slice(slit.ystart-1,slit.ystart+slit.ysize-1),slice(xstart-1,xstart+slit.xsize-1))
    sim = im[slc]
    serr = err[slc]
    sbim = bim[slc]
    sberr = berr[slc]
    # Measure the scaling factor
    goodpix = ((sim/serr>5) & (slim/slerr>5) & slpmask)
    if goodpix.sum()==0:
        goodpix = ((sim/serr>3) & (slim/slerr>3) & slpmask)
    if goodpix.sum()==0:
        print('No good pixels in common to CAL and RATE image')
        return None,None
    scale = np.nanmedian(slim[goodpix]/sim[goodpix])
    print('scale = ',scale)
    sim *= scale
    serr *= scale
    sbim *= scale
    sberr *= scale
    ytrace0 = slytrace.copy()
    mask0 = slpmask.copy()
    ystart = slit.ystart
    ysize = slit.ysize
    
    # Trace is close to the edge, use larger range
    if ((ny-1)-np.max(slytrace) < 4) or (np.min(slytrace)<4):
        if ((ny-1)-np.max(slytrace) < 4):
            nextend = int(np.ceil(4-((ny-1)-np.max(slytrace))))
            ystart = slit.ystart
            ysize = slit.ysize+nextend
            print('Extending the subimage '+str(nextend)+' pixels at the TOP')
            # Extend the wavelength array
            wave = np.vstack((wave,np.zeros([nextend,nx])+np.nan))
            for i in range(nx):
                wave[ny:ny+nextend,i] = np.polyval(wcoef[i,:],np.arange(nextend)+ny)
            # Extend the initial mask
            mask0 = np.vstack((mask0,np.zeros([nextend,nx],bool)))
        else:
            nextend = int(np.floor(4-np.min(slytrace)))
            ystart = slit.ystart-nextend
            ysize = slit.ysize+nextend            
            ytrace0 += nextend
            print('Extending the subimage '+str(nextend)+' pixels at the BOTTOM')
            # Extend the wavelength array
            wave = np.vstack((np.zeros([nextend,nx])+np.nan,wave))
            for i in range(nx):
                wave[0:nextend,i] = np.polyval(wcoef[i,:],np.arange(nextend)-nextend)
            # Extend the initial mask
            mask0 = np.vstack((np.zeros([nextend,nx],bool),mask0))
        ny += nextend
        y = np.arange(ny)
        yy = y.reshape(-1,1) + np.zeros(nx).reshape(1,-1)        
        slc = (slice(ystart-1,ystart+ysize-1),slice(xstart-1,xstart+nx-1))
        sim = im[slc]
        serr = err[slc]
        sbim = bim[slc]
        sberr = berr[slc]
        sim *= scale
        serr *= scale
        sbim *= scale
        sberr *= scale  
    
    # Redo the tracing
    tmask = dln.convolve(mask0.astype(float),np.ones((3,3)))
    tmask[tmask>0] /= tmask[tmask>0]   # normalize to 0/1
    tmask = tmask.astype(bool)
    ttab = tracing(sim*tmask,serr,np.median(ytrace0))

    if len(ttab)==0:
        print('No trace points')
        return None,None
    
    try:
        if len(ttab)<3:
            tcoef = np.array([np.median(ttab['y'])])
            tsigcoef = np.array([np.median(ttab['ysig'])])            
        else:
            tcoef = robust.polyfit(ttab['x'],ttab['y'],2)
            tsigcoef = robust.polyfit(ttab['x'],ttab['ysig'],1)            
        ytrace = np.polyval(tcoef,x)
        ysig = np.polyval(tsigcoef,x)
    except:
        print('tracing coefficient problem')
        import pdb; pdb.set_trace()

    # Create the mask
    ybin = 3
    mask = ((yy >= (ytrace-ybin)) & (yy <= (ytrace+ybin)))

    # Get the background from the background image
    if bratehdu is not None:
        temp1 = sbim.copy()
        temp1[~mask] = np.nan
        sky1 = np.nanmedian(temp1,axis=0)
        diff1 = sbim.copy()-sky1.reshape(1,-1)
        sig1 = dln.mad(diff1[mask])
        temp2 = sbim.copy()
        temp2[(~mask) | (np.abs(diff1)>3*sig1)] = np.nan
        sky2 = np.nanmean(temp2,axis=0)
        sky = np.zeros(ny).reshape(-1,1) + sky2.reshape(1,-1)
    else:
        sky = np.zeros((ny,nx))
    sim -= sky   # subtract background
        
    # Build Gaussian PSF model
    y = np.arange(ny)
    yy = y.reshape(-1,1) + np.zeros(nx).reshape(1,-1)
    gpsf = utils.gauss2dbin(yy,np.ones(nx),ytrace,ysig)
    gpsf /= np.sum(gpsf,axis=0)  # normalize
    # Gaussian PSF extraction
    gflux,gfluxerr,sky,skyerr = extract_psf(sim*mask,gpsf,serr,skyfit=True)

    # Number of good pixels per column with good PSF
    ngood = np.sum((gpsf>0.01)*np.isfinite(sim),axis=0)

    # Optimal extraction
    oflux1,ofluxerr1,otrace1,opsf1 = extract_optimal(sim*mask,ytrace,imerr=serr)

    # Fix bad pixels using the optimal PSF
    fixim,fixmask,fixflux,fixfluxerr = fixbadpixels(sim,serr,opsf1)

    # REJECT BAD PIXELS and redo the optimal extraction
    oflux,ofluxerr,otrace,opsf = extract_optimal(fixim*mask,ytrace,imerr=serr)

    # Boxcar extraction of the fixed image
    boxflux = np.nansum(mask*fixim,axis=0)

    # Optimal extraction looks a little bit better than the boxcar extraction
    #  lower scatter
    flux = oflux
    fluxerr = ofluxerr

    # Save some diagnostic plots
    matplotlib.use('Agg')
    fig = plt.figure(figsize=(12,7))
    # Rate data
    medflux = np.nanmedian(oflux)
    vmin = -medflux/3.
    vmax = medflux
    pl.display(sim,xtitle='X',ytitle='Y',vmin=vmin,vmax=vmax,title='RATE DATA - SLIT_ID '+str(slit.source_id))
    pl.oplot(ytrace,c='red')
    plt.savefig(plotbase+'_ratedata.png',bbox_inches='tight')
    # Rate optimal PSF
    pl.display(opsf,xtitle='X',ytitle='Y',vmin=0,vmax=0.5,title='RATE Optimal PSF - SLIT_ID '+str(slit.source_id))
    pl.oplot(ytrace,c='red')
    plt.savefig(plotbase+'_rateopsf.png',bbox_inches='tight')
    # Flux
    plt.clf()
    plt.plot(oflux,label='Optimal',linewidth=2)
    plt.plot(gflux,label='Gaussian')
    plt.plot(boxflux,label='Boxcar',linestyle='dashed')
    plt.xlabel('X')
    plt.ylabel('Flux')
    plt.ylim(-medflux/3.,1.8*medflux)    
    plt.legend()
    plt.savefig(plotbase+'_rateflux.png',bbox_inches='tight')
    matplotlib.use('MacOSX')
    
    # Get the wavelengths
    pmask = (opsf > 0.01)
    #wav = np.nansum(wave*pmask,axis=0)/np.sum(pmask,axis=0) * 1e4  # convert to Angstroms
    wav = np.nansum(wave*opsf,axis=0)/np.sum(opsf*np.isfinite(wave*opsf),axis=0) * 1e4  # convert to Angstroms

    # Should I use boxcar or optimal extraction?
    # I think boxcar is more precise at high S/N (in the absence of outlier pixels)
    # while optimal extraction is better at low-S/N
    
    # Apply slit correction
    srcxpos = slit.source_xpos
    srcypos = slit.source_ypos
    # SLIT correction, srcxpos is source position in slit
    # the slit is 2 pixels wide
    dwave = np.gradient(wav)
    newwav = wav+2*srcxpos*dwave
    print('Applying slit correction: %.2f pixels' % (2*srcxpos))

    # Apply relative flux calibration correction

    # Add the LSF information
    #  we are essentially working in a slit-less spectrograph regime
    #  the LSF is set by the seeing
    wsig = ysig*dwave
    gdw, = np.where(np.isfinite(wav) & np.isfinite(wsig))
    wsigcoef = np.polyfit(wav[gdw],wsig[gdw],1)
        
    # Put it all together
    sp = Spec1D(flux,err=fluxerr,wave=wav,mask=(fluxerr>1e20),instrument='NIRSpec',
                lsfpars=wsigcoef[::-1],lsftype='Gaussian',lsfxtype='wave')
    sp.date = input_model.meta.date
    sp.jd = Time(input_model.meta.date).jd
    sp.ytrace = ytrace
    sp.source_name = slit.source_name
    sp.source_id = slit.source_id
    sp.slitlet_id = slit.slitlet_id
    sp.source_ra = slit.source_ra
    sp.source_dec = slit.source_dec
    sp.xstart = xstart
    sp.xsize = xsize
    sp.ystart = ystart  # might have been extended
    sp.ysize = ysize
    sp.offset = offset
    sp.tcoef = tcoef
    sp.tsigcoef = tsigcoef
    
    return slsp,sp



import os
import numpy as np
from glob import glob
from dlnpyutils import utils as dln,coords
import doppler
from doppler import spec1d
from astropy.table import Table,hstack,Column
from astropy.io import fits
from chronos import isochrone
import shutil
import subprocess
import tempfile
import traceback
import matplotlib
import matplotlib.pyplot as plt
from . import utils

cspeed = 2.99792458e5  # speed of light in km/s

def continuum(sp,order=1):
    return spec1d.continuum(sp,norder=2,perclevel=50.0,binsize=0.2,interp=True)
    #x = np.linspace(-1,1,len(sp.wave))
    #good1 = (~sp.mask)
    #coef1,_ = dln.ladfit(x[good1],sp.flux[good1])
    #coef1 = coef1[::-1]
    #cont1 = np.polyval(coef1,x)    
    #sig = dln.mad(sp.flux/cont1)
    #good = ((np.abs(sp.flux/cont1-1) < 3*sig) & ~sp.mask)
    #if order==1:
    #    coef,_ = dln.ladfit(x[good],sp.flux[good])
    #    coef = coef[::-1]
    #else:   
    #    coef = robust.polyfit(x[good],sp.flux[good],order)
    #cont = np.polyval(coef,x)
    #return cont,coef

def readspec(filename):
    spec = doppler.read(filename)
    # Doppler models only go up to 18000
    wthresh = 17900
    bad = (spec.wave >= wthresh)
    if np.sum(bad)>0:
        # Trim to wave < 18000
        if spec.ndim==1:
            orig = spec.copy()
            gd, = np.where((spec.wave < wthresh) & (spec.wave > 0))
            npix = len(gd)
            spec.flux = np.zeros(npix,float)
            spec.err = np.zeros(npix,float)+1e30
            spec.mask = np.ones(npix,bool)
            spec.wave = np.zeros(npix,float)
            spec.flux = orig.flux[gd]
            spec.err = orig.err[gd]
            spec.wave = orig.wave[gd]
            spec.mask = orig.mask[gd]
            spec.numpix[0] = npix
            if hasattr(spec,'_cont'):
                spec._cont = spec._cont[gd]
        else:
            orig = spec.copy()
            gd0, = np.where((spec.wave[:,0] < wthresh) & (spec.wave[:,0]>0))
            gd1, = np.where((spec.wave[:,1] < wthresh) & (spec.wave[:,1]>0))
            npix = max([len(gd0),len(gd1)])
            spec.flux = np.zeros([npix,2],float)
            spec.err = np.zeros([npix,2],float)+1e30
            spec.mask = np.ones([npix,2],bool)
            spec.wave = np.zeros([npix,2],float)
            if hasattr(spec,'_cont'):
                spec.cont = np.zeros([npix,2],float)
            for i in range(2):
                gd, = np.where((orig.wave[:,i] < wthresh) & (orig.wave[:,i]>0))
                spec.numpix[i] = len(gd)
                spec.flux[0:len(gd),i] = orig.flux[gd,i]
                spec.err[0:len(gd),i] = orig.err[gd,i]
                spec.mask[0:len(gd),i] = orig.mask[gd,i]
                spec.wave[0:len(gd),i] = orig.wave[gd,i]                
                if hasattr(spec,'_cont'):
                    spec._cont[0:len(gd),i] = spec._cont[gd,i]            
    spec._cont = continuum(spec)
    return spec


def doppler_joint(visitfiles,payne=False,clobber=False,verbose=True):
    """ Run Doppler jointly on visit files for a star."""

    visitfile = visitfiles[0]
    
    if payne:
        figfile = visitfile.replace('.fits','_doppler_joint_payne.png')
        outfile = visitfile.replace('.fits','_doppler_joint_payne.fits')
    else:
        figfile = visitfile.replace('.fits','_doppler_joint.png')
        outfile = visitfile.replace('.fits','_doppler_joint.fits')
    if os.path.exists(outfile) and clobber==False:
        print(outfile+' already exists')
        out = Table.read(outfile,1)
        return out
    if os.path.exists(outfile): os.remove(outfile)        

    # Load the files
    spec = []
    for i in range(len(visitfiles)):
        sp = readspec(visitfiles[i])
        spec.append(sp)
    try:
        out = doppler.jointfit(spec,verbose=verbose,payne=payne)
        sumstr, final, model, specmlist = out
        
        # Save the output
        if os.path.exists(outfile): os.remove(outfile)
        # Summary average values
        Table(sumstr).write(outfile)
        # append other items
        hdulist = fits.open(outfile)
        # append final values for each spectrum
        hdu = fits.table_to_hdu(Table(final))
        hdulist.append(hdu)
        # append best model
        # there's a model for each spectrum, each one gets an extension
        for i,m in enumerate(model):
            hdu = fits.PrimaryHDU(m.flux)
            hdulist.append(hdu)
        hdulist.writeto(outfile,overwrite=True)
        hdulist.close()
        
    except:
        out = None        
        traceback.print_exc()
        
    return out


def doppler_visit(visitfile,payne=False,estimates=None,verbose=True,clobber=False):
    """ Run Doppler one visit stellar spectrum."""
    if payne:
        figfile = visitfile.replace('.fits','_doppler_payne.png')
        outfile = visitfile.replace('.fits','_doppler_payne.fits')
    else:
        figfile = visitfile.replace('.fits','_doppler.png')
        outfile = visitfile.replace('.fits','_doppler.fits')
    if os.path.exists(visitfile) and os.path.getsize(visitfile)==0:
        print(visitfile+' is empty')
        return None,None,None
    if os.path.exists(outfile) and clobber==False:
        print(outfile+' already exists')
        out = Table.read(outfile,1)
        return out
    if os.path.exists(outfile): os.remove(outfile)        

    spec = readspec(visitfile)
    try:
        out,model,specm = doppler.fit(spec,estimates=estimates,verbose=verbose,
                                      figfile=figfile,payne=payne)
        out = Table(out)
        out['snr'] = spec.snr        
        # Save the output to a file
        Table(out).write(outfile)
        # append best model
        hdulist = fits.open(outfile)
        hdu = fits.PrimaryHDU(model.flux)
        hdulist.append(hdu)
        hdulist.writeto(outfile,overwrite=True)
        hdulist.close()
    except:
        out = None
        traceback.print_exc()
        
    return out


def doppler_stack(stackfile,payne=False,estimates=None,verbose=True,clobber=False):
    """ Run Doppler on the stacked stellar spectrum."""
    if payne:
        figfile = stackfile.replace('.fits','_doppler_payne.png')
        outfile = stackfile.replace('.fits','_doppler_payne.fits')
    else:
        figfile = stackfile.replace('.fits','_doppler.png')
        outfile = stackfile.replace('.fits','_doppler.fits')        
    if os.path.exists(outfile) and clobber==False:
        print(outfile+' already exists')
        out = Table.read(outfile,1)
        return out
    if os.path.exists(outfile): os.remove(outfile)  

    if os.path.getsize(stackfile)==0:
        print(stackfile+' is empty')
        return None    
    spec = readspec(stackfile)
    try:
        out,model,specm = doppler.fit(spec,estimates=estimates,verbose=verbose,
                                      figfile=figfile,payne=payne)
        out = Table(out)
        out['snr'] = spec.snr
        # Save the output to a file
        Table(out).write(outfile)
        # append best model
        hdulist = fits.open(outfile)
        hdu = fits.PrimaryHDU(model.flux)
        hdulist.append(hdu)
        hdulist.writeto(outfile,overwrite=True)
        hdulist.close()

    except:
        out = None
        traceback.print_exc()
        
    return out

def run_doppler(obsid,redtag='red',targfile=None,photfile=None,clobber=False,payne=False):
    """ Run Doppler on all the stars."""
    # obsid 'G140H-F100LP-M71_test68'
    
    print('Running doppler for '+obsid)
    
    # Get visit and stack files
    visitfiles = np.char.array(glob(obsid+'/spVisit-*'+redtag+'.fits'))
    visitfiles = np.char.array([v for v in visitfiles if os.path.getsize(v)>0])
    visitbases = [os.path.basename(v) for v in visitfiles if os.path.getsize(v)>0]
    vstarids = [v[8:] for v in visitbases]
    vstarids = [v[:v.find('_jw')] for v in vstarids]
    vstarids = np.char.array(vstarids)
    nvisits = len(visitfiles)

    # Get the list of Stacked files    
    stackfiles = np.char.array(glob(obsid+'/stack/spStack-*_'+redtag+'.fits*'))
    starids = [os.path.basename(c)[8:-len(redtag)-6] for c in stackfiles if os.path.getsize(c)>0]
    starids = np.char.array(starids)
    nstars = len(starids)
    print(nstars,' stars')

    # Loop over the stars
    slist = {}
    vlist = {}
    for i in range(nstars):
        starid = starids[i]
        print(i+1,starid)
        # Visit files
        vind, = np.where(vstarids==starid)
        vfiles = visitfiles[vind]
        #vout = doppler_visit(vfiles)
        # Loop over visit files
        for j in range(len(vfiles)):
            vout = doppler_visit(vfiles[j],clobber=clobber)
            vlist[vfiles[j]]= vout
            if payne:
                estimates = {'TEFF':vout['teff'][0],'LOGG':vout['logg'][0],'FE_H':vout['feh'][0],'RV':vout['vrel'][0]}
                vout_payne = doppler_visit(vfiles[j],clobber=clobber,payne=True,estimates=estimates)
                
        # Run doppler with joint on visits
        if len(vind)>1:
            jout = doppler_joint(vfiles,clobber=clobber)
        
        # Stack file
        sind, = np.where(starids==starid)
        sfiles = stackfiles[sind[0]]
        sout = doppler_stack(sfiles,clobber=clobber)
        slist[starid] = sout
        if payne:
            estimates = {'TEFF':sout['teff'][0],'LOGG':sout['logg'][0],'FE_H':sout['feh'][0],'RV':sout['vrel'][0]}
            sout_payne = doppler_stack(sfiles,clobber=clobber,payne=True,estimates=estimates)        

    # Create visit catalog
    dt = [('starid',str,50),('visitfile',str,200),('id',str,50),('vhelio',float),('vrel',float),('vrelerr',float),
          ('teff',float),('tefferr',float),('logg',float),('loggerr',float),('feh',float),('feherr',float),
          ('chisq',float),('bc',float),('snr',float),('success',bool)]
    vtab = np.zeros(len(vlist),dtype=np.dtype(dt))
    for i,k in enumerate(vlist.keys()):
        vout = vlist[k]
        # G140H-F100LP-M31-FINAL-LONG/spVisit-2609_268389_jw02609009001_04107_00001_red.fits
        vfile = os.path.basename(k)
        starid = k[8:]
        starid = '_'.join(starid.split('_')[0:2])
        vtab['starid'][i] = starid
        vtab['id'][i] = starid.split('_')[1]            
        vtab['visitfile'][i] = vfile
        if vout is None:
            vtab['success'][i] = False
        else:
            for n in vout.dtype.names:
                if n in vout.columns:
                    vtab[n][i] = vout[n][0]
            vtab['success'][i] = True
    vtab = Table(vtab)
            
    # Create stack catalog
    dt = [('starid',str,50),('id',str,50),('vhelio',float),('vrel',float),('vrelerr',float),('teff',float),('tefferr',float),
          ('logg',float),('loggerr',float),('feh',float),('feherr',float),('chisq',float),('bc',float),('snr',float),('success',bool)]
    tab = np.zeros(len(slist),dtype=np.dtype(dt))
    for i,k in enumerate(slist.keys()):
        out = slist[k]
        if out is None:
            tab['starid'][i] = k
            tab['success'][i] = False
        else:
            for n in out.dtype.names:
                if n in out.columns:
                    tab[n][i] = out[n][0]
            tab['starid'][i] = k
            tab['id'][i] = k.split('_')[1]
            tab['success'][i] = True
    tab = Table(tab)


    # Add targeting information
    if targfile is not None:
        targs = Table.read(targfile)
        for c in targs.colnames: targs[c].name = c.lower()
        if 'id' not in targs.columns:
            targs['id'] = np.arange(len(targs))+2  # APT has first ID as 2

        ind1,ind2 = dln.match(tab['id'],targs['id'])
        print(len(ind1),' matches to targeting catalog')
        # Add the new columns
        if len(ind1)>0:        
            for k in targs.columns:
                if k not in tab.columns:
                    tab.add_column(Column(name=k,dtype=targs[k].dtype,length=len(tab)))
                    tab[k][ind1] = targs[k][ind2]

    # Add other catalog information
    if photfile is not None:
        # Match up with catalog that has photometry and other information
        phot = Table.read(photfile)
        for c in phot.colnames: phot[c].name = c.lower()

        ind1,ind2,dist = coords.xmatch(tab['ra'],tab['dec'],phot['ra'],phot['dec'],0.5)
        print(len(ind1),' matches to photometry catalog')
        # Add the new columns
        if len(ind1)>0:        
            for k in phot.columns:
                if k not in tab.columns:
                    tab.add_column(Column(name=k,dtype=phot[k].dtype,length=len(tab)))
                    tab[k][ind1] = phot[k][ind2]
                    
    return tab,vtab


def run_ferre(files,vrel,inter=3,algor=1,init=1,indini=None,nruns=1,cont=1,ncont=0,
              errbar=1,grid='jwstgiant4.dat',save=False,plotsdir=None,plots=True):
    """ 
    Run FERRE on list of spectra

    Parameters
    ----------
    files : list
       Input list of spectrum files.
    vrel : list
       Input list of doppler shifts for the spectra.
    inter : int, optional
       Interpolation algorithm.
         0   nearest neighbor
         1   linear
         2   quadratic Bezier
         3   cubic Bezier
         4   cubic spline
       Default is 3.
    algor : int, optional
       Search algorithm?
        -1   weighted average over the grid  likely
         0   pixel with lowest value         minlocus
         1   Nelder-mead method (Miller's)   minim
         2   BTR method (Csendes/Miller)     global
         3   uobyqa method (Powell/Miller)    uobyqa
         4   truncated Newton                lmqnbc
         5   MCMC                             mcmcde
       Default is 1.
    init : int, optional
       Starting point for search.
         0   use the values in pfile
         1   follow the rules set by keyword indini
       Default is 1.
    indini : list, optional
       Individual control of starting points for each parameter.
    nruns : int, optional
       Number of searches to be done.  Default is 1.
    cont : int, optional
       Continuum parameter:
         1   polynomial fitting  (ncont order, 0-constant, 1-linear)
         2   segmented normaleization  (ncont segments)
         3   running mean  (ncont pixels)
    ncont : int, optional
       Number for continuum normalization.  If cont=1, then ncont
         gives the polynomial order.  If cont=2, then ncont is the
         number of segements.  If cont=3, then ncont is the number
         of pixels for the running mean.
    errbar : int, optional
       Choice of algorithm to compute error bars.
         0   adopt the distance from solution at which chisq=min(chisq)+1
         1   invert the curvature matrix
       Default is 1.
    save : bool, optional
       Do not delete the temporary directory.  Default is save=False.

    Returns
    -------
    tab : table
       Output table of values.
    info : list
       List of dictionaries, one per spectrum, that includes spectra and arrays.

    Example
    -------

    tab,info = run_ferre(files,vrel,algor=1,nruns=5)

    """

    #if type(files) != list:
    #    files = [files]
    #if type(vrel) != list:
    #    vrel = [vrel]
    nfiles = len(files)

    # Loop over the files
    slist = []
    for i in range(nfiles):
        filename = files[i]
        vrel1 = vrel[i]
        if filename is None:
            continue
        if os.path.exists(filename)==False:
            print(filename+' not found')
            continue
        if os.path.getsize(filename)==0:
            print(filename+' is empty')
            continue
        spec = doppler.read(filename)
        spec.vrel = vrel1
        #spec.normalize()
        slist1 = {'filename':filename,'vrel':vrel1,'snr':spec.snr,'spec':spec}
        slist.append(slist1)

    if len(slist)==0:
        print('No spectra to fit')
        return None,None
        
    gridfile = '/Users/nidever/synspec/winter2017/jwst/'+grid
    ferre = '/Users/nidever/projects/ferre/bin/ferre.x'

    # Set up temporary directory
    tmpdir = tempfile.mkdtemp(prefix='ferre')
    curdir = os.path.abspath(os.curdir)
    if plots and plotsdir is None:
        plotsdir = curdir+'/plots/'
    os.chdir(tmpdir)
    print('Running FERRE in temporary directory '+tmpdir)

    # Parameters
    print('INTER = ',inter)
    print('ALGOR = ',algor)
    print('INIT = ',init)
    print('INDINI = ',indini)
    print('NRUNS = ',nruns)
    print('CONT = ',cont)
    print('NCONT = ',ncont)
    print('ERRBAR = ',errbar)
    
    # Create fitting input file
    gridbase = os.path.basename(gridfile)    
    os.symlink(gridfile,tmpdir+'/'+gridbase)
    os.symlink(gridfile.replace('.dat','.unf'),tmpdir+'/'+gridbase.replace('.dat','.unf'))
    os.symlink(gridfile.replace('.dat','.hdr'),tmpdir+'/'+gridbase.replace('.dat','.hdr'))
    lines = []
    lines += ["&LISTA"]
    lines += ["NDIM = 4"]
    lines += ["NOV = 4"]
    lines += ["INDV = 1 2 3 4"]
    lines += ["SYNTHFILE(1) = '"+gridbase+"'"]
    lines += ["F_FORMAT = 1"]
    lines += ["INTER = "+str(inter)]    # cubic Bezier interpolation
    lines += ["ALGOR = "+str(algor)]    # 1-Nelder-Mead,5-MCMC
    lines += ["INIT = "+str(init)]
    if indini is not None:
        lines += ['INDINI = '+np.array2string(np.array(indini)).strip('[]')]        
        nruns = 1
        for term in indini:
            nruns = nruns * term
    #    lines += ["INDINI = "+str(indini)]
    if nruns is not None:
        lines += ["NRUNS = "+str(nruns)]
    lines += ["PFILE = 'ferre.ipf'"]
    lines += ["FFILE = 'ferre.frd'"]
    lines += ["ERFILE = 'ferre.err'"]
    lines += ["WFILE = 'ferre.wav'"]    
    lines += ["OPFILE = 'ferre.opf'"]    # output best-fit parameters and uncertainties
    lines += ["OFFILE = 'ferre.mdl'"]    # output best-fit models
    lines += ["SFFILE = 'ferre.nrd'"]    # normalized data
    lines += ["NOBJ = "+str(len(slist))]
    lines += ["CONT = "+str(cont)]      # Running mean normalization
    lines += ["NCONT = "+str(ncont)]     # Npixel for running mean
    lines += ["WINTER = 2"]    # wavelength interpolate the model fluxes
    lines += ["ERRBAR = "+str(errbar)]
    lines += ["/"]
    dln.writelines('input.nml',lines)

    # Prepare the spectra for input to FERRE
    for i in range(len(slist)):
        slist1 = slist[i]
        filename = slist1['filename']
        spec = slist1['spec']        
        print(i+1,filename,spec.vrel)
        # Synspec spectra are in AIR, JWST spectra are in VACUUM, convert
        spec.wavevac = False
        flux = np.array([])
        err = np.array([])
        wave = np.array([])
        mask = np.array([])
        for sp in spec:
            # trim edge pixels
            logood = np.where(sp.mask==False)[0][0]
            higood = np.where(sp.mask==False)[0][-1]
            slc = slice(logood+30,higood-30)
            flux = np.concatenate((flux,sp.flux[slc]))
            err = np.concatenate((err,sp.err[slc]))
            wave = np.concatenate((wave,sp.wave[slc]))
            mask = np.concatenate((mask,sp.mask[slc]))
        # Remove Doppler shift
        wave /= (1+spec.vrel/cspeed)
        # Trim wavelengths that are too large
        gd, = np.where(wave < 17990)
        flux = flux[gd]
        err = err[gd]
        wave = wave[gd]
        mask = mask[gd]
        # Grow bad mask by ~20 pixels
        oldmask = mask.copy()
        mask = np.convolve(mask,np.ones(61),mode='same').astype(bool)
        # Interpolate over masked pixels
        xpix = np.arange(len(flux))
        bad, = np.where((mask==True) | (flux<=0))
        good, = np.where((mask==False) & (flux>0))
        if len(bad)>0:
            smfilt = utils.nanmedfilt(flux[good],201)
            newpix = dln.interp(xpix[good],smfilt,xpix[bad],kind='linear')
            flux[bad] = newpix
            #flux[bad] = 0.0
            err[bad] = 1e10

        # Put all the information into slist
        slist1['npix'] = len(flux)
        slist1['flux'] = flux
        slist1['err'] = err
        slist1['wave'] = wave
        
    # Put star with maximum npix first
    #  FERRE uses this to set the maximum pixels for ALL the spectra
    npixall = [s['npix'] for s in slist]
    maxind = np.argmax(npixall)
    npix = np.max(npixall)
    if maxind != 0:
        slist1 = slist.pop(maxind)
        slist = [slist1] + slist

    # Make the list of lines to write out for FERRE
    flines,elines,wlines,plines = [],[],[],[]
    for i in range(len(slist)):
        # Need to buffer the arrays to have consistent number of pixels
        flux = slist[i]['flux']
        err = slist[i]['err']
        wave = slist[i]['wave']        
        if len(flux)<npix:
            nmissing = npix-len(flux)
            flux = np.hstack((flux,np.zeros(nmissing)))
            err = np.hstack((err,np.zeros(nmissing)+1e10))
            wave = np.hstack((wave,np.zeros(nmissing))) 
            
        # Convert to ASCII for FERRE
        obs = ''.join(['{:14.5E}'.format(f) for f in flux])
        obserr = ''.join(['{:14.5E}'.format(e) for e in err])
        obswave = ''.join(['{:14.5E}'.format(w) for w in wave])        
        slist[i]['ferre_fline'] = obs
        slist[i]['ferre_eline'] = obserr
        slist[i]['ferre_wline'] = obswave     
        slist[i]['ferre_pline'] = 'spec'+str(i+1)+' 4000.0  2.5  -0.5  0.1'
        slist[i]['ferre_id'] = 'spec'+str(i+1)
        
        flines.append(slist[i]['ferre_fline'])
        elines.append(slist[i]['ferre_eline'])
        wlines.append(slist[i]['ferre_wline'])
        plines.append(slist[i]['ferre_pline'])
        
    dln.writelines('ferre.frd',flines)
    dln.writelines('ferre.err',elines)
    dln.writelines('ferre.wav',wlines)    
    dln.writelines('ferre.ipf',plines)
    
    # Run FERRE
    if os.path.exists('ferre.mdl'): os.remove('ferre.mdl')
    if os.path.exists('ferre.opf'): os.remove('ferre.opf') 
    print('Running FERRE on '+str(len(slist))+' spectra')
    try:
        fout = open('ferre.log','w')
        out = subprocess.call([ferre],stdout=fout,stderr=subprocess.STDOUT)
        fout.close()
        loglines = dln.readlines('ferre.log')
        #out = subprocess.check_output([ferre])
    except:
        fout.close()        
        #traceback.print_exc()
        pass

    
    # Read the output
    olines = dln.readlines('ferre.opf')
    ids = np.char.array([o.split()[0] for o in olines])
    mlines = dln.readlines('ferre.mdl')
    slines = dln.readlines('ferre.nrd')
    nolines = len(olines)
    print(str(nolines)+' lines in output files')

    # Sometimes the number of lines in opf and mdl/nrd do not match

    # Loop over slist
    for i in range(len(slist)):
        slist1 = slist[i]
        filename = slist1['filename']
        ferre_id = slist1['ferre_id']
        
        # Find the right output line for this star
        ind, = np.where(ids==ferre_id)
        if len(ind)==0:
            print(ferre_id+' not found in output file')
            slist1['success'] = False            
            slist1['pars'] = None
            slist1['parerr'] = None
            slist1['snr2'] = None
            slist1['rchisq'] = None
            slist1['model'] = None
            slist1['smflux'] = None
            continue
        ind = ind[0]
        olines1 = olines[ind]
        mlines1 = mlines[ind]
        slines1 = slines[ind]

        # opfile has name, pars, parerr, fraction of phot data points, log(S/N)^2, log(reduced chisq)
        arr = olines1.split()
        pars = np.array(arr[1:5]).astype(float)
        parerr = np.array(arr[5:9]).astype(float)
        logsnr2 = float(arr[10])
        try:
            snr2 = 10**logsnr2            
            logrchisq = float(arr[11])
            rchisq = 10**logrchisq
        except:
            snr2 = np.nan
            logrchisq = np.nan
            rchisq = np.nan
        slist1['success'] = True            
        slist1['pars'] = pars
        slist1['parerr'] = parerr   
        slist1['snr2'] = snr2
        slist1['rchisq'] = rchisq

        mflux = np.array(mlines1.split()).astype(float)
        smflux = np.array(slines1.split()).astype(float)
        # If this star has less pixels than the first star, then we need to trim some pixels
        #  off the end
        if len(mflux) != slist1['npix']:
            mflux = mflux[0:slist1['npix']]
            smflux = smflux[0:slist1['npix']]
        slist1['model'] = mflux
        slist1['smflux'] = smflux

        # Make plots
        if plots:
            if os.path.exists(plotsdir)==False: os.makedirs(plotsdir)
            backend = matplotlib.rcParams['backend']
            matplotlib.use('Agg')
            figsize = 10.0
            fig,ax = plt.subplots()
            fig.set_figheight(figsize*0.5)
            fig.set_figwidth(figsize)
            plt.plot(slist1['wave'],slist1['smflux'],linewidth=0.5)
            plt.plot(slist1['wave'],slist1['model'],linewidth=0.5,alpha=0.8)
            yr = [np.min(slist1['model']),np.max(slist1['model'])]
            yr = [yr[0]-np.ptp(yr)*0.2,yr[1]+np.ptp(yr)*0.2]
            medflux = np.median(slist1['smflux'])
            sigflux = dln.mad(slist1['smflux'])
            yr = [np.min([yr[0],medflux-2.5*sigflux]),np.max([yr[1],medflux+2.5*sigflux])]
            plt.ylim(yr[0],yr[1])
            plt.xlim(np.min(slist1['wave']),np.max(slist1['wave']))
            plt.xlabel('Wavelength (A)')
            plt.ylabel('Normalized Flux')
            plt.title(slist1['filename'])
            plotfile = plotsdir+'/'+os.path.basename(slist1['filename']).replace('.fits','_ferre.pdf')
            plt.savefig(plotfile)
            print('Saving plot to ',plotfile)
            matplotlib.use(backend)  # back to the original backend
            
    # Make a table
    dt = [('filename',str,200),('vrel',float),('snr',float),('teff',float),('tefferr',float),
          ('logg',float),('loggerr',float),('feh',float),('feherr',float),('alpha',float),
          ('alphaerr',float),('pars',float,4),('parerr',float,4),('snr2',float),('rchisq',float)]
    tab = np.zeros(len(slist),dtype=np.dtype(dt))
    for i in range(len(slist)):
        slist1 = slist[i]
        tab['filename'][i] = slist1['filename']
        tab['vrel'][i] = slist1['vrel']
        tab['snr'][i] = slist1['snr']
        if 'pars' in slist1.keys() and slist1['pars'] is not None:
            tab['teff'][i] = slist1['pars'][0]
            tab['tefferr'][i] = slist1['parerr'][0]
            tab['logg'][i] = slist1['pars'][1]
            tab['loggerr'][i] = slist1['parerr'][1]
            tab['feh'][i] = slist1['pars'][2]
            tab['feherr'][i] = slist1['parerr'][2]
            tab['alpha'][i] = slist1['pars'][3]
            tab['alphaerr'][i] = slist1['parerr'][3]
            tab['pars'][i] = slist1['pars']
            tab['parerr'][i] = slist1['parerr']
            tab['snr2'][i] = slist1['snr2']
            tab['rchisq'][i] = slist1['rchisq']
    tab = Table(tab)
    
    # Delete temporary files and directory
    if save==False:    
        os.chdir(curdir)
        shutil.rmtree(tmpdir)

    return tab,slist

def distances(tab,isogrid=None):
    """
    Derives distances from isochrones.
    """

    if isogrid is None:
        print('Loading isochrone grid')
        isotab = Table.read('/Users/nidever/isochrone/parsec_hst/parsec_hst_phot_giants.fits.gz')
        for c in isotab.colnames: isotab[c].name=c.upper()
        isotab['AGE'] = 10**isotab['LOGAGE']
        isotab['METAL'] = isotab['MH']
        isotab['F275WMAG'].name = 'HST_F275WMAG'
        isotab['F336WMAG'].name = 'HST_F336WMAG'
        isotab['F475WMAG'].name = 'HST_F475WMAG'
        isotab['F814WMAG'].name = 'HST_F814WMAG'
        isotab['F110WMAG'].name = 'HST_F110WMAG'
        isotab['F160WMAG'].name = 'HST_F160WMAG'
        isogrid = isochrone.IsoGrid(isotab)
        
    # Each star must have Teff, logg, feh and alpha
    # We'll get distances for a couple ages

    ntab = len(tab)
    dt = [('dist1',float),('dist3',float),('dist5',float),('dist7',float),('dist10',float),('dist12',float),('mndist',float)]
    out = np.zeros(ntab,dtype=np.dtype(dt))
    ages = [1e9, 3e9, 5e9, 7e9, 10e9, 12e9]
    # Loop over the stars
    for i in range(ntab):
        teff = tab['teff'][i]
        logg = tab['logg'][i]
        feh = tab['feh'][i]
        alpha = tab['alpha'][i]        
        # Salaris correction feh
        salfeh = feh + np.log10(0.659*(10**alpha)+0.341)
        usefeh = np.maximum(np.minimum(salfeh,isogrid.maxmetal),isogrid.minmetal)
        # Age loop
        distarr = np.zeros(6)
        for a,age in enumerate(ages):
            # Get the isochrone
            iso = isogrid(age,usefeh,closest=True)

            # EXTINCTION from BEAST
            av = tab['beast_Av'][i]
            rv = tab['beast_Rv'][i]
            iso.ext = av   # extinct the isochrone
            
            data1 = iso.data[0:1]
            for c in iso.colnames:
                data1[c] = dln.interp(iso.data['LOGTE'],iso.data[c],np.log10(teff),kind='quadratic',assume_sorted=False)
            
            # Get distance modulus from f160w and f475w
            agename = 'dist'+str(int(age/1e9))
            obsmags = [tab['f275w'][i], tab['f336w'][i], tab['f475w'][i], tab['f814w'][i], tab['f110w'][i], tab['f160w'][i]]
            obsmags = np.array(obsmags)
            isomags = [data1['HST_F275WMAG'][0], data1['HST_F336WMAG'][0], data1['HST_F475WMAG'][0], data1['HST_F814WMAG'][0],
                       data1['HST_F110WMAG'][0],data1['HST_F160WMAG'][0]]
            isomags = np.array(isomags)
            distmodarr = obsmags - isomags
            good, = np.where((obsmags < 50) & (isomags < 0))
            if len(good)>0:
            #if data1['HST_F160WMAG'][0] < 0:
                #distmodarr = [tab['f275w'][i]-data1['HST_F275WMAG'][0],
                #              tab['f336w'][i]-data1['HST_F336WMAG'][0],
                #              tab['f475w'][i]-data1['HST_F475WMAG'][0],
                #              tab['f814w'][i]-data1['HST_F814WMAG'][0],
                #              tab['f110w'][i]-data1['HST_F110WMAG'][0],
                #              tab['f160w'][i]-data1['HST_F160WMAG'][0]]
                distmod = np.mean(distmodarr[good])
                dist = 10**(1+distmod/5) / 1e3  # kpc
                out[agename][i] = dist
            else:
                out[agename][i] = np.nan
            distarr[a] = out[agename][i]
            
        mndist = np.nanmean(distarr)
        out['mndist'][i] = mndist
        
        fmt = '{:5d} {:9.1f} {:7.3f} {:8.3f} {:8.3f} {:8.3f}  {:9.1f} {:9.1f} {:9.1f} {:9.1f} {:9.1f} {:9.1f}  {:9.1f}'
        print(fmt.format(i+1,teff,logg,feh,alpha,usefeh,*distarr,mndist))
        
    return out
